/* ********** bstack.c - maintain stack of best trees ********** */

#include "lvb.h"

#define CLOSEBR 0100	/* 'close parentheses' */
#define END 00		/* 'terminator (";\n\0")' */
#define LEASTBITS 6	/* number of least significant bits */
#define NUMBER 0300	/* 'number follows' */
#define OPENBR 0200	/* 'open parentheses' */
#define PART1 077	/* least significant bits of number */
#define PART2 (~PART1)	/* most significant bits of number */

/* write byte chr the advance write position pos */
#define sputc(pos, chr) *pos++ = (Byte) (chr), ++written

/* write least sig. 14 bits of i into 2 bytes and adv. write pos. pos by 2 */
#define sputno(pos, i) \
sputc(pos, NUMBER | (i & PART1)), \
sputc(pos, (unsigned) (i & PART2) >> LEASTBITS)

/* unpack integer at read position pos into i & advance pos by 1 */
#define getsno(i, pos) i = *pos & PART1, ++pos, i = i | (*pos << LEASTBITS)

/* get flag bits of byte */
#define flag(chr) (chr & ~PART1)

/* N.B. The 2 flag bits of a byte of a packed tree as read are CLOSEBR, OPENBR,
 * END or NUMBER. NUMBER indicates that a 14-bit object number follows: the
 * rest of the byte contains the LEASTBITS least significant bits, and the next
 * byte contains the most significant bits of the object number. With the other
 * flag bits, the rest of the byte is irrelevant, and nothing should be assumed
 * about the next byte. */

static void dopush(const Branch *const barray, const Branchno root);
static Byte *treepack(Byte *const str, const Branch *const barray,
 const Branchno root);
static Byte *packalloc(void);
static Byte *poppack(void);
static void trunpack(Branch *const barray, const Byte *const packtree);
static void upsize(void);

static Integer trspc = 0;	/* allocation per tree */
static Byte **stack = NULL;	/* stack of packed best trees */
static Lint next = (Lint) 0;	/* next unused element of stack */
static Lint size = (Lint) 0;	/* total trees held in stack */
static Objno outgrp = UNSET;	/* set to og if rooted output desired */
static Boolean rooted = FALSE;	/* tree should be output rooted */

static Byte *treepack(Byte *const str, const Branch *const barray,
const Branchno root)
/* send tree in barray (of root root) to memory pointed to by str,
 * in packed and 'unrooted' form;
 * return new print position in str */
{
	static Byte *s;				/* current write position */
	static Boolean doneabsroot = FALSE;	/* have output root of tree */
	Branchno obj;				/* current object */
	static Integer written = 0;		/* bytes written */

	obj = barray[root].object;
	if (doneabsroot == FALSE)	/* pack whole tree */
	{
		/* start tree */
		s = str;
		sputc(s, OPENBR);
		sputno(s, obj);
		doneabsroot = TRUE;

		treepack(s, barray, barray[root].left);
		treepack(s, barray, barray[root].right);

		/* end tree */
		sputc(s, CLOSEBR);
		sputc(s, END);

		/* check packed tree is expected size (it always will be) */
		if (written != trspc)
			crash("internal error: expected packed tree of %d "
			 "bytes but have written %d bytes", (int) trspc,
			 (int) written);

		/* clean up for next time */
		doneabsroot = FALSE;
		written = 0;
	}
	else	/* pack remainder of tree */
	{
		if (obj != UNSET)	/* leaf */
			sputno(s, obj);
		else
		{
			sputc(s, OPENBR);	/* start subtree */
			treepack(s, barray, barray[root].left);
			treepack(s, barray, barray[root].right);
			sputc(s, CLOSEBR);	/* end subtree */
		}
	}
	return s;

}	/* end treepack() */

Lint bstcnt(void)
/* return number of trees on best stack */
{
	return next;

}	/* end bstcnt() */

static void trunpack(Branch *const barray, const Byte *const packtree)
/* unpack the tree packtree into the array of branches barray, ensuring
 * barray[0] is the root branch */
{
	const Byte *pos = packtree;	/* current position in packed tree */
	Branchno current = 0;		/* current branch number */
	Branchno nextbr = 0;		/* next unused branch number */
	Branchno obj;			/* current object number */
	const Branchno root = 0;	/* root of unpacked trees */

	treeclear(barray);

	/* start tree with root at branch 0 */
	if (*pos != OPENBR)
		crash("internal error: first byte of packed tree\n"
		 "is 0%o but should be 0%o", (int) *pos, (int) OPENBR);
	++pos;
	getsno(obj, pos);
	barray[current].object = obj;
	++pos;
	++nextbr;

	for (; *pos != END; ++pos)
	{
		switch flag(*pos)
		{
		case OPENBR:
			childadd(barray, current, nextbr);
			current = nextbr;
			++nextbr;
			break;
		case CLOSEBR:
			if (current != root)
				current = barray[current].parent;
			break;
		case NUMBER:
			getsno(obj, pos);
			barray[nextbr].object = obj;
			childadd(barray, current, nextbr);
			++nextbr;
			break;
		default:	/* impossible condition */
			crash("internal error: byte 0%o of packed tree has\n"
			 "unknown flag bits", (int) *pos);
                        break;
		}
	}
	if (current != 0)	/* impossible condition */
		crash("internal error: have not returned to root at end of "
		 "unpacking tree");

	return;

}	/* end trunpack() */

void bstinit(Params rcstruct)
/* set things up for later use of best stack, based on matrix and rcstruct */
{
	extern Dataptr matrix;	/* data matrix */

	trspc = matrix->n * 2 - 4;	/* OPENBRs and CLOSEBRs */
	trspc += matrix->n * 2;		/* packed object numbers */
	++trspc;			/* END */

	outgrp = rcstruct.og;
	if (outgrp != UNSET)
		rooted = TRUE;
	else
		rooted = FALSE;
	return;

}	/* end bstinit() */

static Byte *packalloc(void)
/* return pointer to new memory for a packed tree in the best stack */
{

	return (Byte *) alloc((size_t) trspc, "packed tree");

}	/* end packalloc() */

void bstrestore(FILE *const stream, const struct finnegan reparams)
/* add reparams.bst_trees dumped trees to the stack by reading stream;
 * N.B. the best tree stack must have been initialized already by calling
 * bstinit() */
{
	Lint i;			/* loop counter */
	static Branch *barray = NULL;	/* current restored tree */

	if (barray == NULL)	/* 1st call, allocate memory */
	{
		barray = treealloc();
		if (barray == NULL)
			crash("out of memory: cannot allocate for unpacked\n"
			 "current tree to restore best tree stack");
	}
	for (i = 0; i < (Lint) reparams.bst_trees; ++i)
	{
		trundump(stream, barray);
		dopush(barray, 0);	/* trundump() ensures root is 0 */
	}
	return;

}	/* end bstrestore() */

Integer bstpush(const Branch *const barray, const Branchno root)
/* push tree in barray (of root root) on to the best stack only if its
 * topology is not already stored there;
 * return 1 if it is pushed, or 0 if not */
{
	Lint i;					/* loop counter */
	static Branch *stacktree = NULL;	/* current tree on stack */

	if (stacktree == NULL)	/* 1st call, allocate memory */
	{
		stacktree = treealloc();
		if (stacktree == NULL)
			crash("out of memory: cannot allocate for unpacked\n"
			 "current tree in stack");
	}

	/* return before push if not a new topology */
	for (i = 0; i < next; ++i)
	{
		trunpack(stacktree, stack[i]);
		if (treecmp(stacktree, 0, barray, root) == 0)
			return 0;
	}

	/* topology is new so must be pushed */
	dopush(barray, root);
	return 1;

}	/* end bstpush() */

static void dopush(const Branch *const barray, const Branchno root)
/* push tree in barray (of root root) on to best stack, in packed form */
{
	if (next >= size)     /* want more memory for stack */
                upsize();
	treepack(stack[next], barray, root);
        ++next;
        return;
 
}       /* end dopush() */

Lint bstprint(FILE *const outfp)
/* print all trees on stack to file outfp, according to value of static rooted
 * either unrooted or rooted with branches collapsed where necessary;
 * return number of trees printed */
{
	Lint cnt = 0;			/* tree count */
	Boolean *haslen = NULL;		/* rooted: collapse br. where FALSE */
	Branchno root;			/* root of current tree */
	Byte *packtree;			/* current packed tree */
	const Objno d_obj1 = 0;		/* 1st obj. for unrooted trees */
	static Branch *barray = NULL;	/* current unpacked tree */

	if (barray == NULL)	/* 1st call, allocate memory */
	{
		barray = treealloc();
		if (barray == NULL)
			crash("out of memory: cannot allocate for unpacked\n"
			 "current tree to write best trees");
	}
	while ((packtree = poppack()) != NULL)
	{
		trunpack(barray, packtree);
		if (rooted == TRUE)	/* root on outgroup */
		{
			root = objreroot(barray, 0, outgrp);
			haslen = getlmask(barray, root);
		}
		else
			root = objreroot(barray, 0, d_obj1);
		treeprint(outfp, barray, root, haslen);
		++cnt;
	}
	if (fflush(outfp) != 0)
		crash("file write error when writing best trees");
	return cnt;

}	/* end bstprint() */

Lint bstdump(FILE *const outfp)
/* dump all trees on stack to file outfp, with root as first branch (number 0);
 * return number of trees dumped */
{
	Lint cnt = 0;			/* tree count */
	Byte *packtree;			/* current packed tree */
	static Branch *barray = NULL;	/* current unpacked tree */

	if (barray == NULL)     /* 1st call, allocate memory */
	{
		barray = treealloc();
		if (barray == NULL)
			crash("out of memory: cannot allocate for current\n"
			 "unpacked tree for dumping best tree stack");
	}
	while ((packtree = poppack()) != NULL)
	{
		trunpack(barray, packtree);
		treedump(outfp, barray);
		++cnt;
	}
	if (fflush(outfp) != 0)
		crash("file write error when dumping best trees");
	return cnt;

}	/* end bstdump() */

void bstfree(void)
/* free all memory in stack */
{
	Lint i;		/* loop counter */

	for (i = (Lint) 0; i < size; ++i)
		free(stack[i]);
	free(stack);
	next = (Lint) 0;
	size = (Lint) 0;
	stack = NULL;
	return;
 
}       /* end bstfree() */

void bstclear(void)
/* clear stack and reduce its allocation to STACKBLOCK trees if it is
 * greater */
{
	Lint i;		/* loop counter */

	/* resize stack if bigger than 1 block */
	if (size > (Lint) STACKBLOCK)
	{
		/* free trees in stack */
		for (i = (Lint) STACKBLOCK; i < size; ++i)
			free(stack[i]);
		/* resize stack itself */
		stack = (Byte **) realloc(stack, STACKBLOCK * sizeof(Byte *));
		if (stack == NULL)
			crash("memory problem: cannot decrease allocation\n"
			 "for best tree stack from %u to %d trees",
			 (unsigned) size, STACKBLOCK);
		size = STACKBLOCK;
	}

	next = (Lint) 0;	/* clear stack */

}	/* end bstclear() */

static Byte *poppack(void)
/* pop compressed tree from stack in compressed form;
* or return NULL if stack is empty */
{
        --next;
        if (next < (Lint) 0)
        {
                next = (Lint) 0;
                return NULL;
        }
        else
                return stack[next];     /* next push will overwrite */
 
}       /* end poppack() */

static void upsize(void)
/* increase allocation for best tree stack */
{
	Lint i;		/* loop counter */

        size = size + STACKBLOCK;
 
	/* allocate for stack itself */
        if (stack == NULL)   /* 1st call, stack does not exist */
        {
                stack = (Byte **) alloc((size_t) size * sizeof(Byte *),
                 "initial best tree stack");
		next = 0;
        }
        else
        {
        	if (((unsigned long) size * (unsigned long) sizeof(Byte *))
        	 > ((unsigned long) MAX_ALLOC))
        	 	crash("cannot increase allocation for best tree\n"
        	 	 "stack to %u elements because the allocation limit\n"
        	 	 "of %u bytes would be exceeded", (unsigned) size,
        	 	 (unsigned) MAX_ALLOC);
                stack = (Byte **) realloc(stack,
                 (size_t) size * sizeof(Byte *));
                if (stack == NULL)
                        crash("out of memory: cannot increase allocation for\n"
                         "best tree stack to %u elements", (unsigned) size);
        }

	/* allocate for trees within stack */
	for (i = next; i < size; ++i)
		stack[i] = packalloc();

        return;
 
}       /* end upsize() */
