/* ********** datops.c - data matrix functions ********** */

#include "lvb.h"

static void badchar(const Dataptr matrix, const char *const badchars,
 Boolean *const togo, const Boolean verbose);
static void checkmsz(const Lint m, const char *const fnam);
static void checknames(const Dataptr matrix, const char *const fnam);
static void checknsz(const Objno n, const char *const fnam);
static void checktlen(const Integer len, const char *const fnam);
static Charno colcnt(FILE *const stream, const char *const fnam);
static void constchar(const Dataptr matrix, Boolean *const togo,
 const Boolean verbose);
static void cr_ilt(const Objno i, const char *const fnam);
static void cr_mlow(const Charno m, const char *const fnam);
static void cr_nhi(const Objno n, const char *const fnam);
static void cr_nlow(const Objno n, const char *const fnam);
static void cr_nsam(const Dataptr matrix, const char *const fnam,
 const Objno i, const Objno j);
static void cr_obs(const char *const fnam);
static void cr_rne(const char *const fnam, const Objno row, const Charno currm,
 const Charno prevm);
static void cr_sfe(const char *const fnam);
static void cr_tlenhi(const Integer len, const char *const fnam);
static void cr_tnr(const Objno i, const char *const fnam);
static void cr_mhi(const Lint m, const char *const fnam);
static void cr_ntnl(const char *const fnam);
static Charno cutcols(Dataptr matrix, const Boolean *const tocut);
static void cutmsg(const Boolean *const togo, const Charno m,
 const char *const msg);
static void equivchar(Dataptr matrix, char **equivarr);
static void logcut(const Boolean *const cut, const Charno m);
static Dataptr matalloc(const Objno n);
static void relabs(Dataptr matrix, const char matchchar);
static Objno rowcnt(FILE *stream, const char *const fnam);
static char *rowread(FILE *const stream, const char *const fnam,
 const Charno m);
static FILE *skipspace(FILE *const stream, const char *const fnam);
static void testneof(const int c, const char *const fnam);
static void testnforbid(const int c, const char *const fnam);
static void testnueof(const int c, const char *const fnam);
static void testnnz(const int c, const char *const fnam);
static Boolean *tgarralloc(const Integer m);
static char *titleread(FILE *stream, const char *const fnam);

Lint getminlen(const Dataptr matrix)
/* return minimum length of any tree based on matrix */
{
	Lint minlen = 0;	/* return value */
	char *statev;		/* list of states in current character */
	Charno k;		/* loop counter */
	
	for (k = 0; k < matrix->m; ++k)
	{
		statev = getstatev(matrix, k);
		minlen += (Lint) strlen(statev) - (Lint) 1;
	}
	return minlen;

}	/* end getminlen() */

Dataptr matrin(const char *const fnam)
/* return a new matrix read from file called fnam,
 * or crash verbosely on error;
 * push all new memory on to the general memstack, except that for row strings
 * and the array of row strings (since these may be freed by rowfree) */
{
	char *rowstr;		/* current row string */
	char *titlestr;		/* current title string */
	FILE *stream;		/* input file */
	Objno i;		/* loop counter */
	Charno m = UNSET;	/* columns in input matrix */
	Charno currentm;	/* columns in current row of matrix */
	Objno n;		/* rows in input matrix */
	Dataptr mat;		/* new data matrix */

	stream = clnopen(fnam, "r");
	n = rowcnt(stream, fnam);
	checknsz(n, fnam);

	/* read rows and row titles into memory */
	for (i = 0; i < n; ++i)
	{
		titlestr = titleread(stream, fnam);
		if (titlestr == NULL)	/* no row follows title on same line */
			cr_tnr(i, fnam);
		currentm = colcnt(stream, fnam);
		if (currentm != m)
		{
			if (m == UNSET)		/* alright, not got m yet */
			{
				m = currentm;
				mat = matalloc(n);
			}
			else			/* unequal row lengths */
				cr_rne(fnam, i, currentm, m);
		}
		rowstr = rowread(stream, fnam, m);

		/* save strings in matrix data structure */
		mat->row[i] = rowstr;
		mat->rowtitle[i] = titlestr;
	}

	clnclose(stream, fnam);

	/* record dimensions */
	mat->m = m;
	mat->n = n;

	checknames(mat, fnam);
	return mat;

}	/* end matrin() */

void rowfree(Dataptr matrix)
/* free memory used for row strings and array of row strings in matrix,
 * and make the array of row title strings NULL;
 * or, if the array of row title strings is already NULL, do nothing */
{
	Objno i;	/* loop counter */

	if (matrix->row != NULL)
	{
		for(i = 0; i < matrix->n; ++i)
			free(matrix->row[i]);
		free(matrix->row);
		matrix->row = NULL;
	}
	return;

}       /* end rowfree() */

static void checkmsz(const Lint m, const char *const fnam)
/* check m is in the interval [MIN_M, MAX_M] inclusive */
{
	if (m < (Lint) MIN_M)
		cr_mlow(m, fnam);
	else if (m > (Lint) MAX_M)
		cr_mhi(m, fnam);
	return;

}	/* end checkmsz() */

static void checknames(const Dataptr matrix, const char *const fnam)
/* check all row titles in matrix are different and not equal to UNSETNAM,
 * considered case-insensitively;
 * crash verbosely if they are not */
{
	Objno i;	/* loop counter */
	Objno j;	/* loop counter */

	for (i = 0; i < matrix->n; ++i)
	{
		if (strcicmp(matrix->rowtitle[i], UNSETNAM) == 0)
			cr_ilt(i, fnam);
		for (j = 0; j < matrix->n; ++j)
		{
			if ((i != j) && (strcicmp(matrix->rowtitle[i],
			 matrix->rowtitle[j]) == 0))
				cr_nsam(matrix, fnam, i, j);
		}
	}

}	/* end checknames() */

static void checknsz(const Objno n, const char *const fnam)
/* check n is in the interval [MIN_N, MAX_N] */
{
	if (n < (Objno) MIN_N)
		cr_nlow(n, fnam);
	else if (n > (Objno) MAX_N)
		cr_nhi(n, fnam);
	return;

}	/* end checknsz() */

static void checktlen(const Integer len, const char *const fnam)
/* check row title length len in file of name fnam does not exceed the limit
 * MAX_TITLELEN */
{
		if (len > MAX_TITLELEN)
			cr_tlenhi(len, fnam);
		return;

 }	/* end checktlen() */

static Charno colcnt(FILE *const stream, const char *const fnam)
/* count columns of row of matrix proper in stream str (of name fnam);
 * N.B. before return, the read position for stream is restored to its value
 * on calling colcnt */
{
	int inchar = 0;		/* current character of input */
	Lint m = 0;		/* column count */
	fpos_t inipos;		/* file position on function call */

	/* save file position at function call */
	testnnz(fgetpos(stream, &inipos), fnam);

	/* count columns */
	while (inchar != '\n')
	{
		inchar = getc(stream);
		testneof(inchar, fnam);
		if (!(isspace(inchar)))
			++m;
	}

	testnnz(fsetpos(stream, &inipos), fnam);	/* restore file pos. */
	checkmsz(m, fnam);
	return (Charno) m;	

}	/* end colcnt() */

static void cr_tnr(const Objno i, const char *const fnam)
/* crash because 'row' i of matrix in the file called fnam does not have both
 * a title and a row proper */
{
	crash("row %d of matrix in file %s either lacks a title, or has a\n"
	 "title only. Each row must be preceded by a title on the same line,\n"
	 "separated from it by white space.",
	  (int) i + 1, fnam);

}	/* end cr_tnr() */

static void cr_ilt(const Objno i, const char *const fnam)
/* crash because row i of matrix has the illegal title UNSETNAM */
{
	crash("row %d of matrix in file %s has the illegal title '%s'.\n"
	 "Never use this as a title for any row.",
	  (int) i + 1, fnam, UNSETNAM);

}	/* end cr_ilt() */

static void cr_mhi(const Lint m, const char *const fnam)
/* crash because the number of columns in the matrix file called fnam exceeds
 * the upper limit MAX_M */
{
	crash("matrix in file %s has %u columns, which exceeds LVB's upper\n"
	 "limit of %d columns", fnam, (unsigned) m, MAX_M);

 }	/* end cr_mhi() */

static void cr_mlow(const Charno m, const char *const fnam)
/* crash because the number of columns m in the matrix in the file called
 * fnam is less than the lower limit MIN_M */
{
	crash("matrix in file %s has %d columns, which is less than\n"
	 "LVB's lower limit of %d columns.", fnam, (int) m, MIN_M);

}	/* end cr_mlow() */

static void cr_nhi(const Objno n, const char *const fnam)
/* crash because the number of rows n in the matrix in the file called fnam
 * exceeds the upper limit MAX_N */
{
	crash("matrix in file %s has %d rows, which exceeds\n"
	 "LVB's upper limit of %d rows.", fnam, (int) n, MAX_N);

}	/* end cr_nhi() */

static void cr_nlow(const Objno n, const char *const fnam)
/* crash because the number of rows n in the matrix in the file called fnam is
 * less than the lower limit MIN_N */
{
	crash("matrix in file %s has %d rows, which is less than\n"
	 "LVB's lower limit of %d rows.", fnam, (int) n, MIN_N);

}	/* end cr_nlow() */

static void cr_nsam(const Dataptr matrix, const char *const fnam,
 const Objno i, const Objno j)
/* crash because rows i and j of matrix in file named fnam have same title */
{
	crash("row %d and row %d of matrix in file %s have '%s'\n"
	 "and '%s' as their titles. Use a different name for every row.",
	  (int) i + 1, (int) j + 1, fnam, matrix->rowtitle[i],
	  matrix->rowtitle[j]);

}	/* end cr_nsam() */

static void cr_ntnl(const char *const fnam)
/* crash because file of name fnam does not end in '\n' */
{
	crash("final line in file %s is not terminated. The final line must\n"
	 "end in the usual way before end of file.", fnam);

}	/* end cr_ntnl() */

static void cr_obs(const char *const fnam)
/* crash because of obscure error reading matrix in file of name fnam */
{
	crash("error reading matrix in file %s.", fnam);

}	/* end cr_obs() */

static void cr_rne(const char *const fnam, const Objno row, const Charno currm,
 const Charno prevm)
/* crash because row row in file called fnam has currm columns, but previous
 * rows have prevm columns */
{
	crash("matrix in file %s has unequal row lengths: row %d has\n"
	 "%d columns but previous rows have %d columns.", fnam, (int) row + 1,
	 (int) currm, (int) prevm);

}	/* end cr_rne() */

static void cr_sfe(const char *const fnam)
/* crash because of unexpected file error on input stream */
{
	crash("file error on stream for matrix file %s.", fnam);

}	/* end cr_sfe() */

static void cr_tlenhi(const Integer len, const char *const fnam)
/* crash because length of row title, len, exceeds upper limit MAX_TITLELEN */
{
	crash("row title too long (%d machine characters) in file %s:\n"
	 "limit is %d machine characters", (int) len, fnam, MAX_TITLELEN);

}	/* end cr_tlenhi() */

static Dataptr matalloc(const Objno n)
/* return pointer to new matrix, with n pointers to rows and row titles;
 * the strings in these arrays, and the array of state counts, are not
 * allocated for and are initialized to NULL;
 * new memory for matrix struct itself and array of pointers is pushed on to
 * the general memstack */
{
	Objno i;	/* loop counter */
	Dataptr mat;	/* new data matrix */
	char **l_row;	/* new array of rows */

	/* matrix struct itself */
	mat = (Dataptr) alloc(sizeof(struct data), "matrix structure");
	mempush(mat);

	/* array for row title strings */
	mat->rowtitle = (char **) alloc((size_t) (n + 1) * sizeof(char *),
	 "pointers to row title strings");
	mempush(mat->rowtitle);

	/* array for row strings (local variable so matrix->row NULL until can be freed) */
	l_row = (char **) alloc((size_t) (n + 1) * sizeof(char *),
	 "pointers to row strings");

	/* initialize unallocated pointers to NULL */
	for (i = 0; i <= n; ++i)
	{
		mat->rowtitle[i] = NULL;
		l_row[i] = NULL;
	}

	/* initialize scalars to zero */
	mat->m = 0;
	mat->n = 0;

	mat->row = l_row;	/* now can be freed e.g. by rowfree() from cleanup() */

	return mat;

}	/* end matalloc() */

static Objno rowcnt(FILE *stream, const char *const fnam)
/* count rows of matrix in file stream (of name fnam);
 * crash verbosely if last character in file is not '\n';
 * N.B. before return, the read position for stream will be restored to its
 * value on calling rowcnt() */
{
	int inchar;			/* current character of input */
	int oldchar = EOF;		/* previous character of input */
	fpos_t start;			/* initial position */
	Objno n = 0;			/* row count */
	Boolean fulline = FALSE;	/* line contains some non-wht space */

	testnnz(fgetpos(stream, &start), fnam);		/* save file pos. */
	stream = skipspace(stream, fnam);
	while ((inchar = fgetc(stream)) != EOF)
	{
		if (!(isspace(inchar)))
			fulline = TRUE;		/* non-empty line */
		if (inchar == '\n')
		{
			if (fulline == TRUE)
				++n;
			fulline = FALSE;	/* for next line */
		}
		oldchar = inchar;
	}
	if (ferror(stream))	/* 'EOF' above indicated error */
		cr_sfe(fnam);
	if (oldchar != '\n')
		cr_ntnl(fnam);

	testnnz(fsetpos(stream, &start), fnam);		/* rewind */
	return n;

}	/* end rowcnt() */

static char *rowread(FILE *const stream, const char *const fnam,
 const Charno m)
/* read and return m-column row from stream (of filename fnam),
 * converting all cells read to upper case;
 * leave stream so next read gets character after row-terminating newline;
 * N.B. the new string is not pushed on to the general memstack */
{
	char *rowstr;		/* row */
	int inchar;		/* current character of input */
	Charno cells = 0;	/* cells read */

	rowstr = salloc(m, "matrix row strings");

	/* get row text */
	for (;;)
	{
		inchar = getc(stream);
		testnueof(inchar, fnam);
		testnforbid(inchar, fnam);
		if (cells > m)	/* should not happen */
			cr_obs(fnam);
		if (!(isspace(inchar)))	/* is matrix cell */
		{
			rowstr[cells] = (char) inchar;
			++cells;
		}
		else if (inchar == '\n')	/* row terminator */
		{
			if (cells != m)		/* should not happen */
				cr_obs(fnam);
			else			/* have read whole row */
				break;
		}
	}
	
	rowstr[m] = '\0';	/* terminate string */
	supper(rowstr);		/* make upper case */

	return rowstr;

}	/* end rowread() */

static FILE *skipspace(FILE *const stream, const char *const fnam)
/* move to next non-white space character of file stream (of name fnam);
 * return stream, or crash verbosely if EOF encountered */
{
	int inchar = ' ';	/* current character of input */

	/* read up to and including first non-white space byte */
	while (isspace(inchar))
	{
		inchar = getc(stream);
		testnueof(inchar, fnam);
	}

	/* push back the first non-white space byte so it will be read next */
	testneof(ungetc(inchar, stream), fnam);

	return stream;

}	/* end skipspace() */

static void testneof(const int c, const char *const fnam)
/* crash with message about file stream error in file called fnam if c is
 * EOF */
{
	if (c == EOF)
		cr_sfe(fnam);

}	/* end testneof() */

static void testnforbid(const int c, const char *const fnam)
/* crash verbosely if c is a forbidden character, i.e. '\0' or one of the
 * characters in the string FORBIDDEN_CHRS */
{
	if (c == '\0')
		crash("forbidden machine character in file %s: null "
		 "character", fnam);
	if (strchr(FORBIDDEN_CHRS, c) != NULL)

		crash("forbidden machine character in file %s. The following\n"
		 "are not allowed:\n"
		 "%s\n"
		 "and the null character.", fnam, FORBIDDEN_CHRS);

}	/* end testnforbid() */

static void testnueof(const int c, const char *const fnam)
/* crash with message about unexpected end of file in file called fnam if c is
 EOF */
{
	if (c == EOF)
		crash("unexpected end of file when reading file %s.", fnam);

}	/* end testnueof() */

static void testnnz(const int c, const char *const fnam)
/* crash with message about error on file called fnam if c is not zero */
{
	if (c != 0)
		cr_sfe(fnam);

}	/* end testnnz() */

static char *titleread(FILE *stream, const char *const fnam)
/* read a row title from file stream (of name fnam), and return a new string
 * copy of it, or NULL if it is not followed by a matrix row;
 * push the new string on to the general memstack;
 * advance position in stream to start of associated row */
{
	char *title;		/* title string */
	int inchar;		/* current character of input */
	Integer len = 0;	/* length of title string */
	fpos_t start;		/* start of relevant input */

	/* find and save start of row title */
	stream = skipspace(stream, fnam);
	testnnz(fgetpos(stream, &start), fnam);
		
	/* get length of title */
	inchar = 'A';	/* arbitrary non-white space */
	while (!(isspace(inchar)))
	{
		inchar = getc(stream);
		testnueof(inchar, fnam);
		testnforbid(inchar, fnam);
		++len;
	}
	checktlen(len, fnam);

	/* read title into new memory */
	title = salloc(len, "matrix row title strings");
	mempush(title);
	testnnz(fsetpos(stream, &start), fnam);	/* rewind to start of title */
	if (fscanf(stream, "%s", title) != 1)
		cr_sfe(fnam);

	/* move to start of associated row */
	for (inchar = ' '; isspace(inchar); inchar = getc(stream))
	{
		if (inchar == '\n')	/* end of line reached before row */
			return NULL;
	}
	testnueof(inchar, fnam);	/* end of file reached before row */
        testneof(ungetc(inchar, stream), fnam);		/* 1st byte of row */

	return title;

}	/* end titleread() */

static void relabs(Dataptr matrix, const char matchchar)
/* replace all instances of matchchar in matrix with the character state
 * of the first row for the column where it occurs */
{
	Objno i;		/* loop counter */
	Charno k;		/* loop counter */

	for (i = 1; i < matrix->n; ++i)
	{
		for (k = 0; k < matrix->m; ++k)
		{
			if (matrix->row[i][k] == matchchar)
				matrix->row[i][k] = matrix->row[0][k];
		}
	}
	return;

}	/* end relabs() */

static void equivchar(Dataptr matrix, char **equivarr)
/* replace lists of character states in equivarr each with the first
 * member of the list, where they are found in matrix's row strings */
{
	Integer list;		/* current list */
	Integer listcnt;	/* number of lists */
	Objno i;		/* loop counter */
	Charno k;		/* loop counter */

	/* count lists */
	for (listcnt = 0; equivarr[listcnt][0] != '\0'; ++listcnt)
		;

	for (list = 0; list < listcnt; ++list)
	{
		for (i = 0; i < matrix->n; ++i)
		{
			for (k = 0; k < matrix->m; ++k)
			{
				if (strchr(equivarr[list],
				 (int) matrix->row[i][k]) != NULL)
				 	matrix->row[i][k] = equivarr[list][0];
			}
		}
	}
	return;

}	/* end equivchar */

static void badchar(const Dataptr matrix, const char *const badchars,
 Boolean *const togo, const Boolean verbose)
/* make sure matrix->m-element array togo is TRUE where rows contain
 * one of the characters in string badchars;
 * log message about which columns will be ignored if verbose is TRUE */
{
	char *msg;		/* message about columns to be ignored */
	Integer badno;		/* index of current element of badchars */
	Objno i;		/* loop counter */
	Charno k;		/* loop counter */
	Integer badcnt;		/* length of badchars */
	Boolean *tocutl;	/* TRUE where original column must go */

	badcnt = (Integer) strlen(badchars);
	tocutl = tgarralloc(matrix->m);		/* local array of cols to go */

	/* identify columns to cut */
	for (badno = 0; badno < badcnt; ++badno)	/* for ea. 'bad' ch. */
	{
		for (i = 0; i < matrix->n; ++i)		/* for each row */
		{
			for (k = 0; k < matrix->m; ++k)		/* ea. col. */
			{
				if (matrix->row[i][k] == badchars[badno])
				{
					togo[k] = TRUE;		/* caller's */
					tocutl[k] = TRUE;	/* local */
				}	
			}
		}
	}

	if (verbose == TRUE)
	{
		msg = (char *) salloc(50 + badcnt,
		 "message about ignored columns");
		strcpy(msg, "Ignoring columns containing any of: ");
		strcat(msg, badchars);
		cutmsg(tocutl, matrix->m, msg);
		free(msg);	/* local 'automatic' memory */
	}
	free(tocutl);		/* 'automatic' memory */
	return;

}	/* end badchar() */

static void constchar(const Dataptr matrix, Boolean *const togo,
 const Boolean verbose)
/* make sure matrix->m-element array togo is TRUE where matrix column
 * contains only one character state;
 * log details of new columns to ignore if verbose is TRUE */
{
	Charno k;		/* loop counter */
	Objno i;		/* loop counter */
	Boolean *isconst;	/* TRUE where column is constant */

	/* local array of columns to go */
	isconst = tgarralloc(matrix->m);
	for (k = 0; k < matrix->m; ++k)
		isconst[k] = TRUE;

	/* discover variable columns */
	for (i = 1; i < matrix->n; ++i)
	{
		for (k = 0; k < matrix->m; ++k)
		{
			if (matrix->row[i][k] != matrix->row[0][k])
				isconst[k] = FALSE;
		}
	}

	/* update togo, for caller */
	for (k = 0; k < matrix->m; ++k)
	{
		if (isconst[k] == TRUE)
			togo[k] = TRUE;
	}

	if (verbose == TRUE)
		cutmsg(isconst, matrix->m, "Ignoring constant columns");

	free(isconst);	/* 'automatic' memory */
	return;

}	/* end constchar() */

static void cutmsg(const Boolean *const togo, const Charno m,
 const char *const msg)
/* log message that columns for which the m-element array togo is TRUE will be
 * ignored */
{
	Charno cutcnt = 0;	/* number of columns being ignored */
	Charno k;		/* loop counter */
	extern FILE *logfp;	/* log file */

	fprintf(logfp, "%s\n", msg);
	for (k = 0; k < m; ++k)
	{
		if (togo[k] == TRUE)
			++cutcnt;
	}
	if (cutcnt == 0)
		fprintf(logfp, "... none found.\n");
	else
		logcut(togo, m);
	return;

}	/* end cutmsg() */

static Boolean *tgarralloc(const Charno m)
/* return pointer to new m-element array of columns to ingore, with each
 * element initialized to FALSE;
 * does not push the new memory onto the general memstack */
{
	Boolean *togo;	/* new array */
	Charno k;	/* loop counter */

	togo = (Boolean *) alloc((size_t) m * sizeof(Boolean),
	 "array of columns to ignore");

	/* initialize all elements to FALSE ('don't ignore') */
	for (k = 0; k < m; ++k)
		togo[k] = FALSE;

	return togo;

}	/* end tgarralloc() */

void matchange(Dataptr matrix, const Params rcstruct, const Boolean verbose)
/* change and remove columns in matrix, partly in response to rcstruct,
 * verbosely or not according to value of verbose */
{
	Boolean *togo;		/* TRUE where column must go */
	Charno colsgone;	/* columns cut */
	extern FILE *logfp;	/* log file */

	togo = tgarralloc(matrix->m);			/* columns to ignore */
	equivchar(matrix, rcstruct.equivalent);		/* equivalent states */

	/* relative states to absolute */
	if (rcstruct.matchchar != '\0')
		relabs(matrix, rcstruct.matchchar);

	/* 'bad' states */
	if (strlen(rcstruct.badstates) != (size_t) 0)
		badchar(matrix, rcstruct.badstates, togo, verbose);

	constchar(matrix, togo, verbose);	/* compuslory cut */

	/* N.B. a function to mark autapomorphic characters for cutting
	 * could be called at this point. The effect would be more noticable
	 * with unrealistically small test matrices than with real data */

	/* cut the cols as indicated, and crash verbosely if too few remain */
	colsgone = cutcols(matrix, togo);	/* make changes to matrix */
	if (matrix->m < MIN_M)
		crash("after columns are ignored, matrix has %d columns,\n"
		 "which is less than LVB's lower limit of %d columns.",
		  (int) matrix->m, MIN_M);
	else
	{
		if (verbose == TRUE)
			fprintf(logfp, "A total of %d columns will be "
			 "ignored\n", (int) colsgone);
	}

	free(togo);	/* 'automatic' memory */

}	/* end matchange() */

static Charno cutcols(Dataptr matrix, const Boolean *const tocut)
/* remove columns in matrix for which the corresponding element of
matrix->m-element array tocut is TRUE, and update matrix->m;
return the number of columns cut */
{
	char **newrow;			/* rows of reduced matrix */
	Objno i;			/* loop counter */
	Charno newm;			/* new number of columns */
	const Charno oldm = matrix->m;	/* old number of columns */
	Charno k;			/* loop counter */
	Charno newk;			/* current column of reduced matrix */

	/* count columns that will not be cut */
	newm = 0;
	for (k = 0; k < matrix->m; ++k)
	{
		if (tocut[k] == FALSE)
			++newm;
	}

	/* make no change if keeping all columns */
	if (newm == matrix->m)
		return 0;

	/* memory for new matrix row array */
	newrow = (char **) alloc((size_t) (matrix->n + 1) * sizeof(char *),
	 "pointers to new row strings");
	for (i = 0; i < matrix->n; ++i)
		newrow[i] = salloc(newm, "new row strings");
	newrow[matrix->n] = NULL;

	newk = 0;
	for (k = 0; k < matrix->m; ++k)	/* for every column */
	{
		if (tocut[k] == FALSE)	/* keep this column */
		{
			for (i = 0; i < matrix->n; ++i)	/* fill for ea. row */
				newrow[i][newk] = matrix->row[i][k];
			++newk;	/* fill next row next time */
		}
	}

	/* trap impossible condition */
	if (newk != newm)
		crash("internal error: expected new matrix to have %d\n"
		 "columns but it apparently has %d\n", (int) newm, (int) newk);

	/* terminate new row strings */
	for (i = 0; i < matrix->n; ++i)
		newrow[i][newk] = '\0';

	/* update matrix structure */
	rowfree(matrix);
	matrix->row = newrow;
	matrix->m = newm;

	return (oldm - newm);

}	/* end cutcols() */

char *getstatev(const Dataptr matrix, const Charno k)
/* return pointer to string containing 1 instance of each character state in
 * column k of matrix, or crash verbosely if more than MAXSTATES states are
 * found;
 * N.B. string is static and will be overwritten by later calls, and is pushed
 * on to the general memstack */
{
	static char *statev = NULL;	/* array of states */
	Integer statec;			/* number of states */
	Objno i;			/* loop counter */

	if (statev == NULL)	/* 1st call, allocate memory */
	{
		statev = salloc(MAXSTATES, "array of states");
		mempush(statev);
	}

	/* clear record of states */
	statev[0] = '\0';
	statec = 0;
	
	/* update record of states for column k */
	for (i = 0; i < matrix->n; ++i)
	{
		if (strchr(statev, (int) matrix->row[i][k]) == NULL)
		/* new state */
		{
			statev[statec++] = matrix->row[i][k];
			if (statec > MAXSTATES)
				crash("number of character states in column\n"
				 "%d of matrix exceeds the maximum of %d.\n"
				 "Use fewer states, or recompile LVB with a\n"
				 "larger unsigned integeral type for Uint in "
				 "lvb.h", (int) k + 1, MAXSTATES);
			statev[statec] = '\0';	/* for strchr() */
		}
	}
	
	return statev;
	
 }	/* end getstatev() */

static void logcut(const Boolean *const cut, const Charno m)
/* log message saying columns for which m-element array cut is TRUE are being
 * cut */
{
	Charno k;			/* loop counter */
	Integer noperln = 0;		/* no. of numbers on current line */
	const Integer max_noperln = 8; 	/* max. numbers written per line */
	Boolean newline = FALSE;	/* last number followed by '\n' */
	extern FILE *logfp;		/* log file */

	fputs("... will ignore column numbers:\n", logfp);

	/* give formatted list of columns to go */
	for (k = 0; k < m; ++k)
	{
		if (cut[k] == TRUE)
		{
			fprintf(logfp, "%d", (int) k + 1);
			++noperln;
			if (noperln == max_noperln)	/* end line */
			{
				noperln = 0;
				putc('\n', logfp);
				newline = TRUE;
			}
			else	/* just put some space on this line */
			{
				putc('\t', logfp);
				newline = FALSE;
			}
		}
	}
	if (newline == FALSE)
		putc('\n', logfp);

	if (fflush(logfp) != 0)
		crash("write error on file %s", LOGFNAM);
	return;

}	/* end logcut() */
