/* ********** epause.c - pause and restart execution ********** */

#include "lvb.h"

static void preamble(FILE *const savefp);
static void restcrash(void);

#define ENDAMBLE '!'	/* end-of-preamble mark */

void epause(const Branch *const x, const Integer run, const Lint blen,
 const Lint proposed, const Lint accepted, const Lint failedcnt, const Real t,
 const Lint n, const Lint iter)
/* write save file for pausing of simulated annealing, log a message and exit
 * normally after freeing the matrix row array and rows;
 * parameters are as in anneal() */
{
	FILE *savefp;		/* save file name */
	extern FILE *logfp;	/* log file */
	extern Dataptr matrix;	/* data matrix */

	clnremove(PAUSEFNAM);	/* get rid of pause semaphore for next time */
	savefp = clnopen(SAVEFNAM, "w");
	preamble(savefp);
	fprintf(logfp, "Pause requested in run %d, restart LVB to continue\n",
	 (int) run);
	fprintf(savefp, "bst_trees %ld\n", (long) bstcnt());
	fprintf(savefp, "run %ld\n", (long) run);
	fprintf(savefp, "blen %ld\n", (long) blen);
	fprintf(savefp, "proposed %ld\n", (long) proposed);
	fprintf(savefp, "accepted %ld\n", (long) accepted);
	fprintf(savefp, "failedcnt %ld\n", (long) failedcnt);
	fprintf(savefp, "t %g\n", (double) t);
	fprintf(savefp, "n %ld\n", (long) n);
	fprintf(savefp, "iter %ld\n", (long) iter);
	fprintf(savefp, "seed %ld\n", (long) randpint((Integer) MIN_MAX_SEED));

	/* current tree */
	fprintf(savefp, "current_tree\n");
	treedump(savefp, x);

	/* best trees */
	fprintf(savefp, "best_trees\n");
	bstdump(savefp);

	clnclose(savefp, SAVEFNAM);
	rowfree(matrix);	/* otherwise row array and contents leaked */
	exit(0);

}	/* end epause() */

struct finnegan restart(void)
/* return scalars, current tree and best tree left in save file;
 * fill bstack with the best trees in the save file */
{
	int inchar;			/* current character input */
	FILE *savefp;			/* save file */
	struct finnegan reparams;	/* saved values */

	savefp = clnopen(SAVEFNAM, "r");

	/* skip preamble */
	while ((inchar = getc(savefp)) != ENDAMBLE)
	{
		if (inchar == EOF)
			restcrash();
	}

	/* general scalars */
	if (fscanf(savefp, " bst_trees %ld", &reparams.bst_trees) != 1)
		restcrash();
	if (fscanf(savefp, " run %ld", &reparams.run) != 1)
		restcrash();
	if (fscanf(savefp, " blen %ld", &reparams.blen) != 1)
		restcrash();
	if (fscanf(savefp, " proposed %ld", &reparams.proposed) != 1)
		restcrash();
	if (fscanf(savefp, " accepted %ld", &reparams.accepted) != 1)
		restcrash();
	if (fscanf(savefp, " failedcnt %ld", &reparams.failedcnt) != 1)
		restcrash();
	if (fscanf(savefp, " t %lg", &reparams.t) != 1)
		restcrash();
	if (fscanf(savefp, " n %ld", &reparams.n) != 1)
		restcrash();
	if (fscanf(savefp, " iter %ld", &reparams.iter) != 1)
		restcrash();
	if (fscanf(savefp, " seed %ld", &reparams.seed) != 1)
		restcrash();

	/* skip string "current_tree" */
	if (fscanf(savefp, "%*s") != 0)
		restcrash();

	reparams.x = treealloc();
	if (reparams.x == NULL)
		crash("out of memory: cannot allocate for current tree on\n"
		 "restart");
	trundump(savefp, reparams.x);

	/* skip string "best_trees" */
	if (fscanf(savefp, "%*s") != 0)
		restcrash();

	bstrestore(savefp, reparams);
	clnclose(savefp, SAVEFNAM);
	clnremove(SAVEFNAM);
	return reparams;

}	/* end restart() */

static void preamble(FILE *const savefp)
/* write preamble, ending in ENDAMBLE then '\n', to save file savefp */
{
	fprintf(savefp, "PAUSED EXECUTION SAVE FILE FOR %s %s %s\n",
	 PROGNAM, VERSION, SUBVERSION);
	fprintf(savefp, "Leave this file, unaltered, where you found it,\n");
	fprintf(savefp, "and LVB will continue where it left off next time\n");
	fprintf(savefp, "it is run in the same directory. If you would\n");
	fprintf(savefp, "prefer LVB to begin again from scratch, delete\n");
	fprintf(savefp, "this file and that will happen instead.\n");
	fprintf(savefp, "%c\n", ENDAMBLE);
	if (fflush(savefp) == EOF)
		crash("file write error on writing save file %s", SAVEFNAM);
	return;

}	/* end preamble() */

static void restcrash(void)
/* crash due to error in save file when scanning during restart */
{
	crash("error on reading save file %s", SAVEFNAM);

}	/* end restcrash() */
