/* ********** fops.c - file operations ********** */

#include "lvb.h"

FILE *clnopen(const char *const nam, const char *const mod)
/* open file of name nam with mode mod, and return it;
 * or crash verbosely on error */
{
	FILE *fp;	/* file */

	fp = fopen(nam, mod);
	if (fp == NULL)
	{
		if (strcmp(mod, "w") == 0)
			crash("cannot create file %s", nam);
		else if (strcmp(mod, "r") == 0)
			crash("cannot open file %s for reading", nam);
		else if (strcmp(mod, "a") == 0)
			crash("cannot open file %s for appending to", nam);
		else	/* rare mode */
			crash("cannot open file %s with mode \"%s\"",
			 nam, mod);
	}
	return fp;

}	/* end clnopen() */

void clnclose(FILE *const fp, const char *const fnam)
/* check file fp, associated with file named fnam, has no errors and close it;
 * crash verbosely on error;
 * or, if fp is NULL, do nothing */
{
	if (fp != NULL)
	{
		if (ferror(fp) != 0)
			crash("file error on file %s", fnam);
		if (fclose(fp) != 0)
			crash("cannot close file %s", fnam);
	}
	return;

}	/* end clnclose() */

void clnremove(const char *const fnam)
/* remove file whose name is in fnam, and print strong if error */
{
	if (remove(fnam) != 0)
		scream("cannot delete file %s", fnam);
	return;

}	/* end clnremove */

char *f2str(FILE *const stream)
/* return string copy of file pointed to by stream,
 * terminated with newline then '\0';
 * the file is treated as text, so newline is always mapped to '\n' */
{
	char *input;					/* input string */
	unsigned long inbytes;				/* bytes for string */
	unsigned long off;				/* position in str. */
	unsigned long offmax = 0UL;			/* length of string */
	const unsigned long maxom = ULONG_MAX - 2;	/* maximum offmax */

	while (getc(stream) != EOF)
	{
		++offmax;
		if (offmax >= maxom)	/* crash while value has meaning */
			crash("input is too long");
	}
	if (ferror(stream) != 0)
		crash("file error on reading file");
	inbytes = offmax + (unsigned long) 2;	/* '\n', '\0' */

	/* check memory required does not exceed maximum for allocation */
	if (inbytes > (unsigned long) MAX_ALLOC)
		crash("cannot allocate for input: allocation limit of\n"
		 "%u bytes exceeded", (unsigned) MAX_ALLOC);

	input = (char *) alloc((size_t) inbytes, "input");

	/* get string */
	rewind(stream);
	for (off = 0; off < offmax; ++off)
		input[off] = (char) getc(stream);
	if (ferror(stream) != 0)
		crash("file error on reading file");
	input[off] = '\n';
	input[off + 1] = '\0';

	return(input);

}	/* end f2str() */
