/* ********** getparam.c - get and set configurable parameters ********** */

#include "lvb.h"

static void check(const Params *const prms);
static void checkchars(FILE *const stream);
static void defaults(Params *const prms);
static void equivarr(char **equiv, char *liststr);
static Params *prmalloc(void);
static void pupdate(Params *const prms);

static const char *const ignore = "\t\n ";	/* valid white spc in input */

Params *getparam(void)
/* return pointer to user-supplied run-time configuration parameters, if any;
 * otherwise, use defaults */
{
	Params *prms;		/* configurable parameters */

	prms = prmalloc();
	defaults(prms);
	pupdate(prms);
	check(prms);
	return prms;

}	/* end getparam() */

static Params *prmalloc(void)
/* return newly-allocated pointer to a Params structure and all its elements;
 * new memory is pushed on to the general memstack */
{
	Params *prms;	/* new structure */
	Integer i;	/* loop counter */

	prms = alloc(sizeof(Params), "configurable parameters");
	mempush(prms);
	prms->hnam = salloc(MAX_HNAMLEN, "heuristic method name");
	mempush(prms->hnam);

	/* equivalent lists are terminated by "" then NULL for added safety */
	prms->equivalent = alloc((size_t)
	 (2 + MAX_EQUIVALENT) * sizeof(char *), "equivalent lists");
	mempush(prms->equivalent);
	for (i = 0; i <= MAX_EQUIVALENT; ++i)
	{
		prms->equivalent[i] = salloc(MAX_EQUIVALENT,
		 "equivalent strings");
		mempush(prms->equivalent[i]);
	}
	prms->equivalent[MAX_EQUIVALENT + 1] = NULL;

	prms->titleprec = salloc(MAX_TITLEPRECLEN, "titleprec string");
	mempush(prms->titleprec);
	return prms;

}	/* end prmalloc() */

static void defaults(Params *const prms)
/* set seed in *prms to unacceptable value, and other parameters to defaults */
{
	strcpy(prms->hnam, D_HNAM);
	prms->badstates = D_BADSTATES;
	prms->matchchar = D_MATCHCHAR;
	prms->equivalent[0][0] = D_EQUIVALENT;
	prms->og = D_OG;
	prms->ognam = D_OGNAM;
	prms->runs = D_RUNS;
        prms->seed = -1;		/* 'impossible' value */
        prms->t0 = D_T0;
	prms->t1 = D_T1;
	prms->titleprec[0] = D_TITLEPREC;
	prms->maxaccept = D_MAXACCEPT;
	prms->maxpropose = D_MAXPROPOSE;
	prms->maxfail = D_MAXFAIL;
	prms->trees = D_TREES;
	prms->verbose = D_VERBOSE;
	return;

}	/* end defaults() */

static void checkchars(FILE *const stream)
/* check '\0' and white space not in static string ignore do not occur in
 * stream;
 * N.B. stream is rewound before return */
{
	int inchar;	/* current character from stream */

	while ((inchar = getc(stream)) != EOF)
	{
		if ((inchar == '\0') || (isspace(inchar)
		 && (strchr(ignore, inchar) == NULL)))
		 	crash("illegal machine character (value %d) in file "
		 	 "%s", inchar, RCFNAM);
	}

	/* check file error since erros are cleared by rewind */
	if (ferror(stream) != 0)
		crash("file error reading file %s", RCFNAM);

	rewind(stream);
	return;

}	/* end checkchars() */

static void pupdate(Params *const prms)
/* update *prms according to user-supplied values */
{
	char *instr;	/* input */
	char *line;	/* current line of input */
	char *name;	/* variable's name in input */
	char *invalue;	/* mashable copy of value */
	char *value;	/* variable's value in input */
	Integer len;	/* string length */
	long lval;	/* actual integral value */
	Real fval;	/* actual real value */
	FILE *rcfp;	/* configuration file */

	rcfp = fopen(RCFNAM, "r");	/* can't use clnopen() here */
	if (rcfp == NULL)
		instr = NULL;
	else
	{
		checkchars(rcfp);
		instr = f2str(rcfp);
	}
	line = instr;

	/* parse input line by line, if it exists */
	while ((instr != NULL) && ((line = nextnonwspc(line)) != NULL))
	{
		name = line;			/* first token on line */
		line = strchr(line, '\n');	/* for next iteration */
		if (name[0] == COMMENT)		/* ignore line */
			continue;

		/* make tokens terminated strings */
		len = (Integer) strcspn(name, ignore);	/* length of name */
		name[len] = '\0';
		invalue = nextnonwspc(name + len + 1);
		if ((invalue >= line) || (invalue == NULL))	/* missing */
			crash("syntax error in configuration file %s: each\n"
			 "variable named must be followed by a value", RCFNAM);
		len = (Integer) strcspn(invalue, ignore);	/* val. len. */
		invalue[len] = '\0';

		/* make local copy of token since '\0' may be put inside it */
		value = salloc(len, "value from configuration file");
		strcpy(value, invalue);

		/* read value */
		if (strcicmp(name, "seed") == 0)	/* followed by int. */
		{
			lval = strtol(value, NULL, 10);
			if ((lval < 0L) || (lval > (long) MAX_SEED))
				crash("illegal value for variable 'seed' in\n"
 				 "file %s: must be an integer from 0 to %d",
				 RCFNAM, MAX_SEED);
			else
				prms->seed = (int) lval;
		}
		else if (strcicmp(name, "runs") == 0)	/* followed by int. */
		{
			lval = strtol(value, NULL, 10);
			if ((lval < 0L) || (lval > (long) (MAX_INTEGER - 1)))
				crash("illegal value for variable 'runs' in\n"
				 "file %s: must be an integer from 1 to %d",
				 RCFNAM, (int) MAX_INTEGER - 1);
			else
				prms->runs = (Integer) lval;
		}
		else if (strcicmp(name, "verbose") == 0)	/* int. */
		{
			lval = strtol(value, NULL, 10);
			if ((lval == 0L) || (lval == 1L))
				prms->verbose = (Integer) lval;
			else
				crash("illegal value for variable 'verbose'\n"
				 "in file %s: must be either 0 or 1", RCFNAM);
		}
		else if (strcicmp(name, "trees") == 0)	/* f. by long int. */
		{
			lval = strtol(value, NULL, 10);
			if ((lval < 0L) || (lval > ((long) MAX_LINT - 1L)))
				crash("illegal value for variable 'trees' in\n"
				 "file %s", RCFNAM);
			else
				prms->trees = (Lint) lval;
		}
		else if (strcicmp(name, "maxaccept") == 0)	/* long int. */
		{
			lval = strtol(value, NULL, 10);
			if ((lval <= 0L) || (lval > ((long) MAX_LINT - 1L)))
				crash("illegal value for variable\n"
				 "'maxaccept' in file %s", RCFNAM);
			else
				prms->maxaccept = (Lint) lval;
                }
		else if (strcicmp(name, "maxpropose") == 0)	/* long int. */
		{
			lval = strtol(value, NULL, 10);
			if ((lval <= 0L) || (lval > ((long) MAX_LINT - 1L)))
				crash("illegal value for variable\n"
				 "'maxpropose' in file %s", RCFNAM);
			else
				prms->maxpropose = (Lint) lval;
                }
		else if (strcicmp(name, "maxfail") == 0)	/* long int. */
		{
			lval = strtol(value, NULL, 10);
			if ((lval < 0L) || (lval > ((long) MAX_LINT - 1L)))
				crash("illegal value for variable 'maxfail'\n"
				 "in file %s", RCFNAM);
			else
				prms->maxfail = (Lint) lval;
                }
		else if (strcicmp(name, "method") == 0)		/*  string */
		{
			if ((strlen(value) > (size_t) MAX_HNAMLEN)
			 || ((strcicmp(prms->hnam, SAHNAM) != 0)
			 && (strcicmp(prms->hnam, HCHNAM) != 0)
			 && (strcicmp(prms->hnam, RWHNAM) != 0)))
				crash("illegal method '%s' in file %s",
				 value, RCFNAM);
			else
				strcpy(prms->hnam, value);
		}
		else if (strcicmp(name, "outgroup") == 0)	/* string */
		{
			prms->ognam = (char *) salloc((Integer) strlen(value),
			 "outgroup name");
			mempush(prms->ognam);
			strcpy(prms->ognam, value);
		}
		else if (strcicmp(name, "ignorechars") == 0)	/* string */
		{
			if (strcicmp(value, UNSETNAM) == 0)
				prms->badstates = "";
			else
			{
				prms->badstates = salloc(
				 (Integer) strlen(value),
				 "list of 'ignore' states");
				mempush(prms->badstates);
				strcpy(prms->badstates, value);
				supper(prms->badstates);
			}
		}
		else if (strcicmp(name, "titleprec") == 0)	/* string */
		{
			if (strcicmp(value, UNSETNAM) == 0)
				prms->titleprec[0] = '\0';
			else	/* N.B. leave case as found */
			{
				if (strlen(value) > (size_t) MAX_TITLEPRECLEN)
					crash("invalid value for titleprec\n"
					 "in file %s: use a string of up to "
					 "%d machine characters, or\n"
					 "'titleprec none' for none", RCFNAM,
					 MAX_TITLEPRECLEN);
				else
					strcpy(prms->titleprec, value);
			}
		}
		else if (strcicmp(name, "equivalent") == 0)	/* str. list */
			equivarr(prms->equivalent, value);
		else if (strcicmp(name, "matchchar") == 0)	/* character */
		{
			if (strlen(value) != (size_t) 1)
			/* BEGIN BUG FIX in LVB 1.0A: allow 'matchchar none' */
			{
				if (strcicmp(value, UNSETNAM) == 0)
					prms->matchchar = '\0';
				else
					crash("invalid value for matchchar\n"
					 "in file %s: should be a single "
					 "machine character, or use\n"
				 	 "'matchchar none' for none", RCFNAM);
			}
			else
			/* END BUG FIX */
				prms->matchchar = (char)
				 toupper((int) value[0]);
		}
		else	/* followed by floating point number */
		{
			fval = (Real) strtod(value, NULL);
			if (errno == ERANGE)
				crash("value for variable '%s' in file %s is\n"
				 "out of range", name, RCFNAM);
			if (strcicmp(name, "t0") == 0)
			{
				prms->t0 = fval;
				if ((prms->t0 <= 0.0) 
				 || (prms->t0 > 1.0))
					crash("illegal value %g for variable "
					 "'t0' in file %s: please use a\n"
					 "number greater than 0.0 and less "
					 "than or equal to 1.0",
					 (double) prms->t0, RCFNAM);
			}
			else if (strcicmp(name, "t1") == 0)
			{
				prms->t1 = fval;
				if ((prms->t1 <= 0.0) 
				 || (prms->t1 > 1.0))
					crash("illegal value %g for variable "
					 "'t1' in file %s: please use a\n"
					 "number greater than 0.0 and less "
					 "than or equal to 1.0",
					 (double) prms->t1, RCFNAM);
			}
			else	/* unknown or inappropriate token */
				crash("unknown variable '%s' in file %s", name,
				 RCFNAM);
		}
		line = invalue + len + 1;	/* go past end of value */
		free(value);			/* alloc'd at start of loop */

		/* crash on extra value(s) on line */
		if ((nextnonwspc(name + len + 1) != NULL)
		 && (nextnonwspc(name + len + 1)
		 < strchr(name + len + 1, '\n')))
			crash("syntax error in configuration file %s:\n"
			 "each variable named must be followed by exactly one "
			 "value", RCFNAM);
	}

	/* free 'automatic' memory */
	if (instr != NULL)
		free(instr);

	/* set default seed if none specified by user */
	if (prms->seed == -1)
	        prms->seed = (int) time(NULL);
        if ((prms->seed < 0) || (prms->seed > MAX_SEED))	/* can't use */
                crash("cannot get random number seed from system clock:\n"
                 "please specify a value from 0 to %d for variable seed in\n"
                 "configuration file %s", MAX_SEED, RCFNAM);

	if (rcfp != NULL)
		clnclose(rcfp, RCFNAM);
	return;

}	/* end pupdate() */

static void equivarr(char **equiv, char *liststr)
/* fill equiv with lists of states named as equivalent to each other in 
 * liststr (in which lists are separated by commas, without white space),
 * converted to upper case;
 * terminate equiv with the string "";
 * N.B. liststr may be mashed by addition of '\0's */
{
	char *currlist;		/* current list string */
	Integer currnum = 0;	/* current list number */

	if (strcicmp(liststr, "none") == 0)
	{
		equiv[0][0] = '\0';
		return;
	}

	while ((currlist = strtok(liststr, ",")) != NULL)
	{
		if (strlen(currlist) > MAX_EQUIVALENT)
			crash("value of equivalent too long in file %s,\n"
			 "no equivalent list may contain more than %d\n"
			 "machine characters (biological character states)",
			  RCFNAM, MAX_EQUIVALENT);
		if (currnum >= MAX_EQUIVALENT)
			crash("too many lists for value of equivalent in\n"
			 "file %s, there may be no more then %d lists",
			 RCFNAM, MAX_EQUIVALENT);
		strcpy(equiv[currnum], currlist);
		supper(equiv[currnum]);
		++currnum;
		liststr = NULL;	/* get second and subsequent lists */
	}
	equiv[currnum][0] = '\0';
	return;

}	/* end equivarr() */

static void check(const Params *const prms)
/* check parameters in prms are mutually sensible, and crash verbosely
 * if not */
{
	if (prms->t1 > prms->t0)
		crash("t1 must be less than or equal to t0 (currently t0 is\n"
		"%g and t1 is %g)", (double) prms->t0, (double) prms->t1);

}	/* end check() */
