/* ********** lvb.h - main header for lvb ********** */

#include <ctype.h>
#include <errno.h>
#include <limits.h>
#include <math.h>
#include <signal.h>
#include <stdarg.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include "myuni.h"

/* the program */
#define PROGNAM "lvb"			/* program file name */
#define VERSION "1.0A"			/* version of program */
#define SUBVERSION "18 August 1997"	/* very precise version of version */

/* value that may have to be decreased to suit some versions of malloc() */
#define MAX_ALLOC (~((size_t) 0))	/* max. bytes contiguous allocation */

/* values some people may feel the urge to change */
#define CHECK_INTERVAL ((Lint) 5000)	/* check stop/pause every ... trees */
#define COMMENT '%'			/* comment line in config. file */
#define FORBIDDEN_CHRS ",:;()[]`'\""	/* not allowed in matrix file */
#define LVB_FNAMSIZE 128		/* maximum bytes for file names */
#define MAX_EQUIVALENT 32		/* max. eq. st. lists & sts per list */
#define STACKBLOCK 16			/* block sz for (re)alloc. of stacks */
#define UNSET (-1)			/* value of integral vars when unset */
#define UNSETNAM "none"			/* value for names if unset */

/* implementation-independent limits */
#define MIN_MAX_SEED 32767	/* min. max. rand. no. seed on any system */
#define MIN_M 1			/* min. no. of characters for any analysis */
#define MAX_N 16383		/* to allow 14-bit packed object numbers */
#define MIN_N 4			/* min. no. of objs, for rearrangeable tree */
#define MAX_TITLEPRECLEN 32	/* max. length of titleprec string */
#define MAX_HNAMLEN 15		/* max. length of method name string */

/* limits that reassure the user */
#define MAX_M 32766		/* max. columns in data matrix */
#define MAX_TITLELEN 120	/* max. length of each object name string */

/* unchangeable types */
typedef unsigned char Byte;		/* byte for packed objects */
typedef enum { FALSE, TRUE } Boolean;	/* boolean digit */

/* types that might be changeable in a later version */
typedef int Branchno;		/* branch no. (array offset, count or UNSET) */
typedef int Charno;		/* biol. char. no. (offset, count or UNSET) */
typedef int Integer;		/* general purpose signed integer */
typedef int Objno;		/* object no. (array offset, count or UNSET) */
typedef double Real;		/* general purpose floating point variable */
typedef unsigned Uint;		/* general purpose unsigned integer */

/* get long signed integral type of at least 32 bits */
#if INT_MAX < 2147483647
typedef long Lint;	/* int not big enough, use long */
#define MAX_LINT ((long) LONG_MAX)
#else
typedef int Lint;	/* int is big enough, use int */
#define MAX_LINT INT_MAX
#endif	/* if INT_MAX < 2147483647 */

/* limits that can vary with system or typedefs */
#define MAX_INTEGER INT_MAX		/* maximum value for Integer */
/* max. states per character */
#define MAXSTATES ((Integer) sizeof(Uint) * CHAR_BIT)

/* matrix and associated information */
typedef struct data
{
        char **row;		/* array of row strings */
        Charno m;		/* number of columns */
        Objno n;		/* number of rows */
        char **rowtitle;	/* array of row title strings */ 
} *Dataptr;

/* branch of tree */
typedef struct
{
	Branchno parent;        /* parent branch number, UNSET in root */
	Branchno left;          /* child 1 number */
	Branchno right;         /* child 2 number */
	Objno object;           /* object number if leaf, otherwise UNSET */
} Branch;

/* configurable parameters */
typedef struct
{
	char *hnam;		/* all: heuristic method name */
	char matchchar;		/* all: indicates 'same state as in 1st row' */
	char *titleprec;	/* all: to precede stats title lines */
	char *badstates;	/* all: ignore cols containing these symbols */
	char **equivalent;	/* all: lists of equivalent character states */
	char *ognam;		/* all: outgroup object's name */
	Objno og;		/* all: outgroup object number */
	int seed;		/* all: seed for random number generator */
	Integer verbose;	/* all: verboseness level */
	Integer runs;		/* all: number of runs */
	Lint trees;		/* H.C., R.W.: changes to consider */
	Real t0;		/* S.A.: initial temperature */
	Real t1;		/* S.A.: second temperature */
	Lint maxaccept;		/* S.A.: max. changes at one temperature */
	Lint maxpropose;	/* S.A.: max. changes considered at a temp. */
	Lint maxfail;		/* S.A.: max. consecutive temps w/ no change */
} Params;

/* restart details */
struct finnegan
{
	long bst_trees;		/* number of best trees */
	long run;		/* current run */
	long blen;		/* best length */
	long proposed;		/* changes proposed at this temperature */
	long accepted;		/* changes accepted at this temperature */
	long failedcnt;		/* consecutive temps with no accepted change */
	long n;			/* ordinal number of this temperature */
	Real t;			/* current temperature */
	long iter;		/* current tree number */
	long seed;		/* random number seed */
	Branch *x;		/* current tree */
};

/* fixed file names */
#define LOCKFNAM "lock"		/* lock message file */
#define LOGFNAM "log"		/* progress log file */
#define MATFNAM "data"		/* matrix file name */
#define PAUSEFNAM "pause"	/* pause execution semaphore file name */
#define RCFNAM "params"		/* run-time configuration parameters file */
#define SAVEFNAM "save"		/* save file generated on pausing */
#define STOPFNAM "stop"		/* stop semaphore file name */

/* file name bases (run-specific suffixes will be used) */
#define LENFNAM "stat"		/* current tree and length file name prefix */
#define RESFNAM "res"		/* results file name prefix */
#define SUMFNAM "sum"		/* summary of trees per run file name */
#define TREE1FNAM "ini"		/* initial tree file name prefix */

/* method names */
#define SAHNAM "anneal"		/* simulated annealing */
#define HCHNAM "hillclimb"	/* hill climbing */
#define RWHNAM "randomwalk"	/* random walk */
#define D_HNAM SAHNAM		/* default method */

/* constants used by all methods */
#define NO_OGNAM UNSETNAM	/* unrooted tree */
#define D_RUNS 1		/* number of runs */
#define D_EQUIVALENT '\0'	/* no state is equivalent to another */
#define D_OG UNSET 		/* outgroup object number */
#define D_OGNAM NO_OGNAM	/* outgroup object name */
#define D_BADSTATES ""		/* use all columns */
#define D_MATCHCHAR '\0'	/* states aren't relative to first row */
#define D_TITLEPREC '\0'	/* to precede all lines of stats titles */
#define D_VERBOSE 1		/* verbose output */

/* default hill climbing and randomwalk parameter */
#define D_TREES ((Lint) 20000)	/* number of trees to examine */

/* default simulated annealing parameters */
#define D_T0 ((Real) 1.0)		/* initial temperature */
#define D_T1 ((Real) 0.9)		/* second temperature */
#define D_MAXACCEPT ((Lint) 50)		/* max. changes at a temperature */
#define D_MAXPROPOSE ((Lint) 5000)	/* max. proposals at a temperature */
#define D_MAXFAIL ((Lint) 2)		/* max. temperatures with no change */

/* data-hiding macro to access integer-encoded matrix */
#define binmat(i, j) tbinmat[j][i]

/* LVB global functions */
void *alloc(const size_t bytes, const char *const msg);
Lint anneal(const Branch *const inittree, Branchno root, const Real t0,
 const Real t1, const Lint maxaccept, const Lint maxpropose,
 const Lint maxfail, const Integer run, FILE *const lenfp, 
 const char *const titleprec, Lint lenbest, Lint proposed,
 Lint accepted, Lint failedcnt, Real t, Lint n, Lint iter);
Branchno brcnt(void);
void bstclear(void);
Lint bstcnt(void);
Lint bstdump(FILE *const outfp);
void bstfree(void);
void bstinit(Params rcstruct);
Lint bstprint(FILE *const outfp);
Integer bstpush(const Branch *const barray, const Branchno root);
void bstrestore(FILE *const stream, const struct finnegan reparams);
Branchno childadd(Branch *const tree, const Branchno destination,
 const Branchno newchild);
void cleanup(void);
void clnclose(FILE *const fp, const char *const fnam);
FILE *clnopen(const char *const nam, const char *const mod);
void clnremove(const char *const fnam);
void crash(const char *const fmt, ...);
void epause(const Branch *const x, const Integer run, const Lint blen,
 const Lint proposed, const Lint accepted, const Lint failedcnt, const Real t,
 const Lint n, const Lint iter);
char *f2str(FILE *const stream);
Boolean *getlmask(const Branch *const tree, const Branchno root);
Lint getminlen(const Dataptr matrix);
Params *getparam(void);
Lint getplen(const Branch *const tree, const Branchno root);
Branchno getroot(const Branch *const barray);
char *getstatev(const Dataptr matrix, const Charno k);
Lint hillclimb(const Branch *const inittree, Branchno root,
 const Lint maxtrees, FILE *const lenfp, const char *const titleprec);
Uint **makebin(const Dataptr matrix);
void matchange(Dataptr matrix, const Params rcstruct, const Boolean verbose);
Dataptr matrin(const char *const fnam);
void memfree(void);
Lint mempush(void *const mem);
Branchno mutate(Branch *const desttree, const Branch *const sourcetree,
 Branchno root);
char *nextnonwspc(const char *string);
Branchno objreroot(Branch *const barray, const Branchno oldroot,
 const Objno newrobj);
Boolean pausenow(void);
void randtree(Branch *const barray);
Lint randomwalk(const Branch *const inittree, Branchno root,
 const Lint maxtrees, FILE *const lenfp, const char *const titleprec);
Integer randpint(const Integer upper);
struct finnegan restart(void);
void rowfree(Dataptr matrix);
char *salloc(const Integer len, const char *const msg);
void scream(const char *const fmt, ...);
void settraps(void);
Branch *treealloc(void);
Boolean stopnow(void);
Integer strcicmp(const char *const s1, const char *const s2);
char *supper(char *const s);
void treeclear(Branch *const barray);
void treecopy(Branch *const dest, const Branch *const source);
Integer treecmp(const Branch *const tree_1, const Branchno root_1,
 const Branch *const tree_2, Branchno root_2);
void treedump(FILE *const stream, const Branch *const tree);
Branch *treepop(void);
void treeprint(FILE *const stream, const Branch *const barray,
 const Branchno root, const Boolean *const haslen);
void treeswap(Branch **const tree1, Branchno *const root1,
 Branch **const tree2, Branchno *const root2);
void trundump(FILE *const stream, Branch *const barray);


