/* ********** main.c - LVB ********** */

#include "lvb.h"

/* external variables for the whole program */
FILE *logfp = NULL;		/* log file */
Dataptr matrix = NULL;		/* data matrix */
Uint **tbinmat = NULL;		/* transposed, integer-encoded matrix */
Branchno nbranches = 0;		/* number of branches in each tree */

static void getsoln(const Params rcstruct, struct finnegan *reparams);
static void logstim(void);
static void logtree1(const Branch *const barray, const Integer run,
 const Boolean isrestart);
static void setlock(void);
static void smessg(const Boolean isrestart, const Integer run);
static void writeinf(Params prms);
static Params setog(Params rcstruct);

main()
{
	const char *logmod;		/* log file opening mode */
	FILE *savefp;			/* save file with restart info. */
	Params rcstruct;		/* configurable parameters */
	Boolean rstflag;		/* execution is a restart */
	Boolean verbstup;		/* verbose start-up */
	Boolean manual_cleanup = TRUE;	/* call cleanup() manually at end */
	struct finnegan reparams;	/* restart details */

	setlock();
	if (atexit(cleanup) == 0)
		manual_cleanup = FALSE;

	/* decide if this is a restart according to presence of save file */
	if ((savefp = fopen(SAVEFNAM, "r")) != NULL)	/* is restart */
	{
		rstflag = TRUE;
		verbstup = FALSE;
		logmod = "a";
		clnclose(savefp, SAVEFNAM);
	}
	else				/* isn't restart */
	{
		rstflag = FALSE;
		verbstup = TRUE;
		logmod = "w";
	}

	/* global files: part 1 */
	logfp = fopen(LOGFNAM, logmod);		/* can't use clnopen() here */
	if (logfp == NULL)
	{
		fprintf(stderr, "FATAL ERROR: %s: cannot open log file\n",
		 PROGNAM);
		exit(1);
	}
	/* BEGIN BUG FIX in LVB 1.0A: to work under Windows (part 1 of 2) */
/*	clnclose(stdin, "stdin");	*/	/* no standard input */
	/* END BUG FIX */

	settraps();
	rcstruct = *(getparam());

	/* entitle log file */
	fprintf(logfp, "This is %s version %s %s\n", PROGNAM, VERSION,
	 SUBVERSION);

	/* global files: part 2 */
	/* BEGIN BUG FIX in LVB 1.0A: to work under Windows (part 2 of 2) */
/*	clnclose(stderr, "stderr");	*/
/*	clnclose(stdout, "stdout");	*/
	/* END BUG FIX */

	matrix = matrin(MATFNAM);	/* get data matrix */
	nbranches = brcnt();
	rcstruct = setog(rcstruct);	/* get outgroup object number */

	bstinit(rcstruct);		/* set up best tree stack */
	if (rstflag == FALSE)	/* echo configurable information */
		writeinf(rcstruct);
	else			/* get restart info. */
		reparams = restart();
	matchange(matrix, rcstruct, verbstup);	/* cut columns from matrix */

	tbinmat = makebin(matrix);		/* set up encoded matrix */
	fprintf(logfp, "Minimum theoretical tree length for this matrix is "
	"%ld\n", (long) getminlen(matrix));
	if (rstflag == FALSE)	/* not a restart */
	{
		logstim();
		rinit(rcstruct.seed);
		getsoln(rcstruct, NULL);
	}
	else			/* is a restart */
	{
		fprintf(logfp, "Resuming\n");
		logstim();
		fprintf(logfp, "Seed for random number generator (based on "
		 "last random number) = %d\n", (int) reparams.seed);
		rinit((int) reparams.seed);
		getsoln(rcstruct, &reparams);
	}
	rowfree(matrix);
	if (manual_cleanup == TRUE)
		cleanup();
	exit(0);

}	/* end main() */

static void setlock(void)
/* lock directory, or crash with message to stderr if lock already present;
 * N.B. only ever writes to stderr */
{
	FILE *lockfp;	/* lock file */

	lockfp = fopen(LOCKFNAM, "r");	/* can't use clnopen() here */
	if (lockfp != NULL)	/* directory is locked */
	{
		fprintf(stderr, "%s: FATAL ERROR: ", PROGNAM);
		fprintf(stderr, "%s apparently already running in ", PROGNAM);
		fprintf(stderr, "current directory.\n");
		fprintf(stderr, "if not, delete file %s and restart\n",
		 LOCKFNAM);
		exit(1);
	}

	/* set lock: make a file of name LOCKFNAM containing 1 space only */
	lockfp = fopen(LOCKFNAM, "w");
	if (lockfp == NULL)
	{
		fprintf(stderr, "FATAL ERROR: %s: cannot create lock file\n",
		 PROGNAM);
		exit(1);
	}
	if (fputc(' ', lockfp) == EOF)
	{
		fprintf(stderr, "FATAL ERROR: %s: cannot write to lock file\n",
		 PROGNAM);
		exit(1);
	}
	fclose(lockfp);		/* don't check value, failure is not serious */

	return;

}	/* end setlock() */

static void logstim(void)
/* log start time with message, if possible */
{
	time_t stim;	/* time */

	stim = time(NULL);
	if (stim != -1)		/* time is available */
		fprintf(logfp, "Start time: %s", ctime(&stim));
	return;

}	/* end logstim() */

static void writeinf(Params prms)
/* write initial details to log file */
{
	Integer i;	/* loop counter */

	/* heuristic method details */
        fprintf(logfp, "Search method used, method = %s\n", prms.hnam);
        if (strcicmp(prms.hnam, SAHNAM) == 0)
        {
                fprintf(logfp, "Initial temperature, t0 = %g\n",
                 (double) prms.t0);
		fprintf(logfp, "Second temperature, t1 = %g\n",
		 (double) prms.t1);
		fprintf(logfp, "Maximum changes accepted before "
		 "decreasing temperature, maxaccept = %ld\n",
		 (long) prms.maxaccept);
		fprintf(logfp, "Maximum changes considered before "
		 "decreasing temperature, maxpropose = %ld\n",
		 (long) prms.maxpropose);
		fprintf(logfp, "Maximum consecutive temperatures with no "
		 "change accepted before\n"
		 "ending cooling, maxfail = %ld\n", (long) prms.maxfail);
        }
        else if ((strcicmp(prms.hnam, HCHNAM) == 0)
         || (strcicmp(prms.hnam, RWHNAM) == 0))
                fprintf(logfp, "Number of trees to examine, trees = %ld\n",
                 (long) prms.trees);
        else    /* impossible condition */
                crash("internal error: unknown method '%s'", prms.hnam);

	/* universally applicable details */
        fprintf(logfp, "Seed for random number generator, seed = %d\n",
         prms.seed);
        fprintf(logfp, "Runs, runs = %d\n", (int) prms.runs);
	fprintf(logfp, "Verboseness, verbose = %d\n", (int) prms.verbose);
	fprintf(logfp, "Outgroup, outgroup = %s", prms.ognam);
	if (strcicmp(prms.ognam, NO_OGNAM) == 0)
		fprintf(logfp, " (will output unrooted trees)");
	putc('\n', logfp);
	fprintf(logfp, "'Same as row 1' state, matchchar = ");
	if (prms.matchchar == '\0')
		fputs("none\n", logfp);
	else
		fprintf(logfp, "%c\n", prms.matchchar);
	if (strlen(prms.badstates) > 0)
		fprintf(logfp, "Will ignore columns containing any of: %s\n",
		 prms.badstates);
	else
		fprintf(logfp, "Not requested to ignore any columns\n");
	for (i = 0; prms.equivalent[i][0] != '\0'; ++i)
	{
		fprintf(logfp, "Equivalent character states, equivalent ");
		if (prms.equivalent[1][0] != '\0')	/* > 1 list */
			fprintf(logfp, "(%d) ", (int) i + 1);
		fprintf(logfp, "= %s\n", prms.equivalent[i]);
	}
	if (i == 0)
		fputs("Equivalent character states, equivalent = none\n",
		 logfp);
	fprintf(logfp, "String to precede titles in statistics files, "
	 "titleprec = ");
	if (prms.titleprec[0] == '\0')
		fprintf(logfp, "none\n");
	else
		fprintf(logfp, "%s\n", prms.titleprec);	
	fprintf(logfp, "Matrix has %d rows and %d columns\n", (int) matrix->n,
	 (int) matrix->m);
	if (fflush(logfp) == EOF)
		crash("write error on log file %s", LOGFNAM);
	return;

}	/* end writeinf() */

static Params setog(Params rcstruct)
/* return new version of rcstruct, with outgroup name and number set */
{
	Objno i;	/* loop counter */

	if (strcicmp(rcstruct.ognam, D_OGNAM) == 0)	/* use default og */
                rcstruct.og = D_OG;
        else                            /* user has given the name */
        {
        	rcstruct.og = UNSET;
                for (i = 0; i < matrix->n; ++i)
                {
                        if (strcicmp(matrix->rowtitle[i], rcstruct.ognam) == 0)
                                rcstruct.og = i;
                }
                if (rcstruct.og == UNSET)       /* name not found */
                        crash("unknown object '%s' named as outgroup in file\n"
                         "%s", rcstruct.ognam, RCFNAM);
        }
	return rcstruct;

}	/* end setog() */

static void logtree1(const Branch *const barray, const Integer run,
 const Boolean isrestart)
/* log initial tree for run run (in barray) to outfp, using isrestart to
 * determine if this is a restart and choosing the output file name
 * accordingly */
{
        char outfnam[LVB_FNAMSIZE];    	/* current file name */
	const char ressfx[] = "r";	/* restart file suffix */
	const Branchno root = 0;	/* after randtree() or trundump() */
	FILE *outfp;			/* output file */

	/* choose filename according to whether this is a restart or not */
	if (isrestart == TRUE)
		sprintf(outfnam, "%s%d%s", TREE1FNAM, (int) run, ressfx);
	else
		sprintf(outfnam, "%s%d", TREE1FNAM, (int) run);

	/* create tree file */
	outfp = clnopen(outfnam, "w");
	treeprint(outfp, barray, root, NULL);
	clnclose(outfp, outfnam);

}	/* end logtree1() */

static void getsoln(const Params rcstruct, struct finnegan *reparams)
/* get and output solution(s) according to parameters in rcstruct, if reparams
 * is NULL;
 * if reparams is not NULL, use the restart details it contains as well as
 * parameters in rcstruct for a restart */
{
	char fnam[LVB_FNAMSIZE];	/* current file name */
	const char *smod;		/* mode for opening statistics files */
	Lint blen;			/* shortest length */
	Integer run;			/* current run number */
	Lint treec;			/* number of trees found */
	Lint treelength;		/* length of each tree found */
	const Branchno initroot = 0;	/* initial root of all trees */
	Lint accepted;			/* changes accepted at current temp. */
	Lint failedcnt;			/* 'failed' temperatures */
	Lint iter;			/* ordinal number of current tree */
	Lint n;				/* ordinal number of current temp. */
	Lint proposed;			/* changes proposed at current temp. */
	Real t;				/* current temperature */
	FILE *sumfp;			/* best length file */
	FILE *lenfp;			/* current length file */
	FILE *resfp;			/* results file */
	Branch *tree;			/* initial tree */
	Boolean stopflag = FALSE;	/* stopped by user */
	Boolean isrestart;		/* this is a restart */

	if (reparams != NULL)	/* is a restart */
	{
		smod = "a";
		isrestart = TRUE;
		run = (Integer) reparams->run;
		blen = (Lint) reparams->blen;
		proposed = (Lint) reparams->proposed;
		accepted = (Lint) reparams->accepted;
		failedcnt = (Lint) reparams->failedcnt;
		t = (Real) reparams->t;
		n = (Lint) reparams->n;
		iter = (Lint) reparams->iter;
		tree = reparams->x;
	}
	else			/* not a restart */
	{
		smod = "w";
		isrestart = FALSE;
		run = 0;
		blen = (Lint) UNSET;
		proposed = (Lint) UNSET;
		accepted = (Lint) UNSET;
		failedcnt = (Lint) UNSET;
		t = (Real) UNSET;
		n = (Lint) UNSET;
		iter = 0;
		tree = treealloc();
		if (tree == NULL)
			crash("out of memory: cannot allocate for initial "
			 "tree");
	}

	/* open and entitle statistics file shared by all runs */
	sumfp = clnopen(SUMFNAM, smod);
	if (isrestart == FALSE)
	{
		if (rcstruct.titleprec[0] != '\0')
			fprintf(sumfp, "%s ", rcstruct.titleprec);
		fprintf(sumfp, "Run\tInitLen\tBestLen\tTrees\n");
	}
        for (; run < rcstruct.runs; ++run)
        {
		smessg(isrestart, run);

		/* open statistics output files */
		sprintf(fnam, "%s%d", LENFNAM, (int) run);
		if (rcstruct.verbose > 0)
			lenfp = clnopen(fnam, smod);
		else
			lenfp = NULL;

		if (isrestart == FALSE)		/* is start of run */
		{
			randtree(tree);
			bstclear();
	                /* start run's entry in sum file */
	                fprintf(sumfp, "%d\t%ld\t", (int) run,
	                 (long) getplen(tree, 0));
		}
		logtree1(tree, run, isrestart);

		/* find solution(s) */
		if (strcicmp(rcstruct.hnam, SAHNAM) == 0)
			treelength = anneal(tree, initroot, rcstruct.t0,
			 rcstruct.t1, rcstruct.maxaccept, rcstruct.maxpropose,
			 rcstruct.maxfail, run, lenfp, rcstruct.titleprec,
			 blen, proposed, accepted, failedcnt, t, n, iter);
		else if (strcicmp(rcstruct.hnam, HCHNAM) == 0)
			treelength = hillclimb(tree, initroot,
			 rcstruct.trees, lenfp, rcstruct.titleprec);
		else if (strcicmp(rcstruct.hnam, RWHNAM) == 0)
			treelength = randomwalk(tree, initroot,
			 rcstruct.trees, lenfp, rcstruct.titleprec);
		else
			crash("internal error: unknown method in params");

		if (treelength < 0)	/* stopped by user */
		{
			treelength = -(treelength);
			stopflag = TRUE;
		}

		/* close closable statistics files */
		sprintf(fnam, "%s%d", LENFNAM, (int) run);
		if (lenfp != NULL)
			clnclose(lenfp, fnam);

		/* output solution and its details */
                sprintf(fnam, "%s%d", RESFNAM, (int) run);
                resfp = clnopen(fnam, "w");
                treec = bstprint(resfp);
                clnclose(resfp, fnam);
		fprintf(sumfp, "%ld\t%ld\n", (long) treelength, (long) treec);

		/* won't use length summary file until end of next run */
		if (fflush(sumfp) == EOF)
			crash("file write error on file %s", SUMFNAM);

		if (stopflag == TRUE)
			fprintf(logfp, "Early stop requested in run %d\n",
			 (int) run);
                fprintf(logfp, "Ending run %d\n", (int) run);
		if (fflush(logfp) != 0)
			crash("write error on log file");
		if (stopflag == TRUE)
		{
			clnremove(STOPFNAM);
			break;
		}
		
		/* start next run as normal: it is never a restart */
		reparams = NULL;
		isrestart = FALSE;
        }
	clnclose(sumfp, SUMFNAM);
	return;

}	/* end getsoln() */

static void smessg(const Boolean isrestart, const Integer run)
/* print start or continuation of run message as appropriate */
{
	static Boolean normmsg = FALSE;		/* must give normal message */
	extern FILE *logfp;			/* log file */

	/* give appropriate message */
	if (normmsg == FALSE)
	{
		if (isrestart == TRUE)
		{
			fprintf(logfp, "Continuing run %d\n",
			 (int) run);
		}
		else
			fprintf(logfp, "Beginning run %d\n", (int) run);
	}
	else
		fprintf(logfp, "Beginning run %d\n", (int) run);

	if ((fflush(logfp) != 0) || (ferror(logfp) != 0))
			crash("file error on file %s", LOGFNAM);
	normmsg = TRUE;
	return;

}	/* end smessg() */
