/* ********** makebin.c - make integer-encoded matrix ********** */

#include "lvb.h"

static Uint **bmatalloc(const Dataptr matrix);

Uint **makebin(const Dataptr matrix)
/* return an integer-encoded mapping of the matrix proper, where each state in
 * each column (biological character) is represented by a different power of 2;
 * this can be used to calculate state sets using bitwise operations OR
 * (intersection) and AND (union);
 * the matrix returned is transposed with respect to the original, since it
 * will be accessed by biological characters, so it should only be accessed by
 * the macro binmat() (defined in lvb.h) to avoid confusion;
 * new memory for the matrix is pushed on to the general memstack */
{
	char *statev;			/* state array for current character */
	Uint **ltbinmat;		/* integer-encoded mapping of matrix */
	Objno i;			/* loop counter */
	Charno k;			/* loop counter */
	ptrdiff_t power;		/* power of 2 for current cell */
	
	ltbinmat = bmatalloc(matrix);
	for (k = 0; k < matrix->m; ++k)
	{
		statev = getstatev(matrix, k);
                for (i = 0; i < matrix->n; ++i)
		{
			power = strchr(statev, (int) matrix->row[i][k])
			 - statev;
			if (power < 0)	/* impossible condition */
				crash("internal error: cannot create encoded\n"
				 "version of character %d", (int) k + 1);
			/* N.B. don't rotate by 0: it is not wholly explicit
			 * in K. & R. ed. 2 that rotating by 0 is allowed */
			else if (power > 0)
				ltbinmat[k][i] = 1U << power;
			else
				ltbinmat[k][i] = 1U;
		}
	}

	return ltbinmat;

}	/* end makebin() */

static Uint **bmatalloc(const Dataptr matrix)
/* return memory for a new integer-encoded, transposed version of matrix->row;
 * push new memory on to the general memstack;
 * N.B. the new matrix rows are contiguous if possible, otherwise allocated
 * individually */
{
	Charno k;			/* loop counter */
	Uint **bmat;			/* new encoded matrix */
	unsigned long matspace;		/* bytes of contig. memory for cells */

	/* allocate for first dimension, allowing for terminating NULL */
	bmat = (Uint **) alloc((size_t) (matrix->m + 1) * sizeof(Uint *),
	 "encoded matrix");
	mempush(bmat);

	/* allocate for rows all at once if possible, otherwise individually */
	matspace = (unsigned long) matrix->m * (unsigned long) matrix->n
	 * (unsigned long) sizeof(Uint);
	if ((matspace > (unsigned long) MAX_ALLOC)	/* can't allocate */
	 || ((bmat[0] =
	 (Uint *) malloc((size_t) matspace)) == NULL))	/* failed allocation */
	{
		for (k = 0; k < matrix->m; ++k)
		{
			bmat[k] = (Uint *) alloc((size_t) matrix->n,
			 "encoded matrix row");
			mempush(bmat[k]);
		}
	}
	else	/* have allocated for rows contigously above */
	{
		mempush(bmat[0]);
		for (k = 1; k < matrix->m; ++k)
			/* N.B. can't assume Integer can hold m * n */
			bmat[k] = bmat[0] + (size_t) matrix->n * (size_t) k;
	}

	bmat[matrix->m] = NULL;		/* however the rows were allocated */
	return bmat;

}	/* end bmatalloc() */
