/* ********** mem.c - maintain a memory stack to prevent leaks ********** */

#include "lvb.h"

static void memchange(const Lint pos, void *const newptr);
static void *mempop(void);
static void memsize(void);

static void **memstack = NULL;	/* pointers to memory to free before exit */
static Lint memstsize = 0;	/* no. of elements allocated for in memstack */
static Lint memstnext = 0;	/* next unused element in memstack */

/* N.B. Once pointers have been pushed on to the memstack, they can only be
 * popped then freed one after the other until the stack is empty, by calling
 * memfree(). */

Lint mempush(void *const mem)
/* push pointer to memory on to stack to be freed at exit;
return pointer's index in stack array */
{
        if (memstnext >= memstsize)     /* want more memory for stack */
                memsize();
        memstack[memstnext++] = mem;
        return (memstnext - 1);

}       /* end mempush() */

static void memchange(const Lint pos, void *const newptr)
/* change entry in memstack whose offset from stack start is pos, to newptr */
{
	memstack[pos] = newptr;
	return;
	
}	/* end memchange() */

void memfree(void)
/* free all pointers in memstack in reverse order of their registration with
 * mempush */
{
	void *tofree;	/* memory to free */

	while ((tofree = mempop()) != NULL)
		free(tofree);
	memstack = NULL;	/* freed, since was as bottom of memstack */
	return;

}	/* end memfree() */

static void *mempop(void)
/* pop pointer to memory from stack;
or return NULL if stack is empty */
{
	--memstnext;
	if (memstnext < 0)
	{
		memstnext = 0;
		return NULL;
	}
	else
		return memstack[memstnext];	/* next push will overwite */

}	/* end mempop() */

static void memsize(void)
/* increase allocation for memstack by STACKBLOCK elements */
{
	memstsize += (Lint) STACKBLOCK;		/* new stack size */
	if (memstack == NULL)	/* 1st call, stack does not exist */
	{
		memstack = (void **) alloc((size_t) memstsize * sizeof(void *),
		 "general memstack");
		mempush(memstack);	/* to free stack itself, last of all */
		memstnext = 1;
	}
	else
	{
		memstack = (void **)
		 realloc(memstack, (size_t) memstsize * sizeof(void *));
		if (memstack == NULL)
			crash("out of memory: cannot increase allocation for\n"
			 "general memstack to %d elements", (int) memstsize);
		memchange((Lint) 0, memstack);	/* in case stack has moved */
	}
	return;

}	/* end memsize() */
