/* ********** parsim.c - tree evaluation functions ********** */

#include "lvb.h"

static void cr_otdne(const Integer els);
static void makessets(Boolean *const haslen, const Branchno lower,
 const Branchno upper);
static Lint ssupdate (void);
static void statalloc(void);
static void uppass(const Branchno root);
static void upupdate(const Branchno parent, const Branchno child);

static Uint *stateset = NULL;		/* branches' state sets */
static const Branch *barray;		/* tree */
static Branchno *todoarr = NULL;	/* branches of unknown stateset */
static Branchno ext_todo;		/* elements in todoarr when full */

Boolean *getlmask(const Branch *const tree, const Branchno root)
/* return pointer to static array whose elements are TRUE if the
 * corresponding branch may have length > 0 and FALSE otherwise, according to
 * the accelerated transformation model of character change;
 * N.B. caller should not free returned array, which will be overwritten by
 * subsequent calls and will be freed transparently on exit */
{
	Integer off = 0;		/* array index */
	Objno obj;			/* current object */
	Branchno i;			/* loop counter */
	Charno k;			/* current character number */
	static Boolean *haslen = NULL;	/* TRUE if evidence for branch */
	extern Uint **tbinmat;		/* transposed, encoded matrix */
	extern Branchno nbranches;	/* branches in tree */
	extern Dataptr matrix;		/* data matrix */

	/* set up static objects */
	barray = tree;
        if (!stateset)	/* first call, allocate memory */
        	statalloc();
	if (haslen == NULL)	/* first call, allocate memory */
	{
		haslen = (Boolean *)
		 alloc((size_t) nbranches * sizeof(Boolean), "length mask");
		mempush(haslen);
	}

	/* clear length mask */
	for (i = 0; i < nbranches; ++i)
		haslen[i] = FALSE;

	/* get l. mask for all branches except root (which has UNSET parent) */
	for (k = 0; k < matrix->m; ++k)		/* for each character */
	{
		off = 0;
		/* set state sets for non-root leaves */
		for (i = 0; i < nbranches; ++i)
		{
			
			if ((obj = barray[i].object) != UNSET)	/* leaf */
				stateset[i] = binmat(obj, k);
			else
			{
				stateset[i] = 0U;
				todoarr[off++] = i;
			}
		}

		/* get root's state set from children */
		stateset[root] = 0U;
		todoarr[off++] = root;

		/* check impossible condition */
		if (off != (Integer) ext_todo)
			cr_otdne(off);

		/* get MPR state sets for character k */
		ssupdate();		/* downpass */
		uppass(root);

		/* update haslen according to state sets for this character */
		makessets(haslen, 0, root - 1);
		makessets(haslen, root + 1, nbranches - 1);
	}

	haslen[root] = TRUE;	/* since tree is rooted */
	return haslen;

}	/* end getlmask() */

static void makessets(Boolean *const haslen, const Branchno lower,
 const Branchno upper)
/* ensure element numbers lower to upper inclusive in haslen are TRUE where
 * static stateset indicates the branch of that number may have length */
{
	Branchno i;	/* loop counter */

	for (i = lower; i <= upper; ++i)
	{
		if (stateset[i] != stateset[barray[i].parent])
			haslen[i] = TRUE;
	}
	return;

}	/* end makessets() */

Lint getplen(const Branch *const tree, const Branchno root)
/* return total changes in all characters for tree in barray (of root root);
 * N.B. all characters in one function) to guarantee moderate speed, but it is
 * well worth inlining ssupdate() as well */
{
	Integer off;			/* array index */
	Objno obj;			/* current object */
	Charno k;			/* current character number */
	register Branchno i;		/* loop counter (heavily used) */
	Lint changes = 0;		/* number of changes in character k */
	extern Uint **tbinmat;		/* transposed encoded matrix */
	extern Branchno nbranches;	/* branches in tree */
	extern Dataptr matrix;		/* data matrix */

	/* set up static objects */
	barray = tree;
        if (!stateset)	/* first call, allocate memory */
		statalloc();

	/* calculate total tree length */
	for (k = 0; k < matrix->m; ++k)
	{
		off = 0;
		/* set state sets for non-root leaves */
		for (i = 0; i < nbranches; ++i)
		{
			if ((obj = barray[i].object) != UNSET)	/* leaf */
				stateset[i] = binmat(obj, k);
			else
			{
				stateset[i] = 0U;
				todoarr[off++] = i;
			}
		}

		/* get root's state set from children */
		stateset[root] = 0U;
		todoarr[off++] = root;

		/* check impossible condition */
		if (off != (Integer) ext_todo)
			cr_otdne(off);

		changes += ssupdate();	/* get unions used to update s. sets */

		/* add one step if root's character isn't in its state set */
		changes += 
		 (Lint) !(stateset[root] & binmat(barray[root].object, k));
	}

	return changes;

}	/* end getplen() */

static void statalloc(void)
/* set ext_todo, and allocate for static arrays stateset and todoarr */
{
	extern Branchno nbranches;	/* branches per tree */
	extern Dataptr matrix;		/* data matrix */

	ext_todo = nbranches - matrix->n + 1;	/* interior branches + root */
	stateset = (Uint *) alloc((size_t) nbranches * sizeof(Uint),
	 "state sets");
	mempush(stateset);
	todoarr = (Branchno *) alloc((size_t) ext_todo * sizeof(Branchno),
	 "'to do' array");
	mempush(todoarr);
	return;

}	/* end statalloc() */

static Lint ssupdate (void)
/* use a down pass to update empty state sets in stateset according to the
 * non-empty ones and the shape of the tree in barray;
 * return the number of union operations necessary to do this;
 * N.B. when compiling for speed, this function should be inlined into
 * getplen() */
{
	Lint unions = 0;		/* no. of union operations */
	Uint leftss;			/* left child's state set */
	Uint rightss;			/* right child's state set */
	Branchno l_todo = ext_todo;	/* modifiable copy of 'to do' count */
	/* N.B. current is used 2 to 5 times per inner loop, and i 3 to 4
	 * times. No other automatic variables are definitely or possibly used
	 * so much, so I have declared them register. I haven't done this for
	 * more variables because I don't want to limit assignment of external
	 * variables to registers by the compiler */
	register Branchno current;	/* current branch */
	register Branchno i;		/* loop counter */

	while (l_todo)
	{
		for (i = 0; i < l_todo; ++i)
		{
			current = todoarr[i];
			if ((rightss = stateset[barray[current].right])
			 && (leftss = stateset[barray[current].left]))
			{
				if (!(stateset[current] = leftss & rightss))
				/* no intersection, get union */
				{
					stateset[current] = leftss | rightss;
					++unions;
				}
				todoarr[i] = todoarr[--l_todo];
			}
		}
	}
	return unions;

}	/* end ssupdate() */

static void uppass(const Branchno root)
/* use an up pass to update state sets in stateset according to the shape of
 * the tree in barray (of root root);
 * use after ssupdate() to leave accelerated transformation state sets in
 * static array stateset */
{
	if (barray[root].left != UNSET)		/* not leaf */
	{
		upupdate(root, barray[root].left);
		upupdate(root, barray[root].right);
		uppass(barray[root].left);
		uppass(barray[root].right);
	}

	return;

}	/* end uppass */

static void upupdate(const Branchno parent, const Branchno child)
/* in an up pass, update child's entry in stateset according to the parent's */
{
	Uint newset;	/* child's new state set */

	newset = stateset[child] & stateset[parent];
	if (!(newset))		/* no intersection, get union */
		stateset[child] = stateset[child] | stateset[parent];
	else			/* use intersection */
		stateset[child] = newset;
	return;

}	/* end upupdate */

static void cr_otdne(const Integer els)
/* crash because of internal error: todoarr has had els elements set, which
 * differs from the number expected, ext_todo */
{
	crash("internal error: todoarr has had %d elements set, but %d\n"
	 "were expected", (int) els, (int) ext_todo);

}	/* end cr_otdne() */
