/* ********** solve.c - solving functions ********** */

#include "lvb.h"

static FILE *lengthfp;		/* length file */
static Branch *x = NULL;	/* current configuration */
static Branch *xdash = NULL;	/* proposed new configuration */

static void checkedom(void);
static void configalloc(void);
static void lenlog(const char *const fmt, ...);

Lint anneal(const Branch *const inittree, Branchno root, const Real t0,
 const Real t1, const Lint maxaccept, const Lint maxpropose,
 const Lint maxfail, const Integer run, FILE *const lenfp,
 const char *const titleprec, Lint lenbest, Lint proposed, Lint accepted,
 Lint failedcnt, Real t, Lint n, Lint iter)
/* seek parsimonious tree from initial tree in inittree (of root root)
 * with initial temperature t0, and subsequent temperatures obtained by
 * multiplying the current temperature by
 * by (t1 / t0) ** n * t0 (where n is the ordinal number of this temperature,
 * originally 0) after at least maxaccept changes have been
 * accepted or maxpropose changes have been proposed, whichever is
 * sooner;
 * return after maxfail consecutive temperatures have led to no new accepted
 * solution;
 * run is the number of the current run (expected to be initially 0), lenfp is
 * for output of current tree length and associated details or NULL for no
 * such output, titleprec is to precede the line of title text in output
 * statistics;
 * with a restart, other parameters contain values of anneal() automatic
 * variables of the same name saved at the time of pausing,
 * otherwise they should all have value (Real) UNSET or (Lint) UNSET */
{
	Real deltah;		/* change in energy (1 - C.I.) */
	Integer deltalen;	/* change in length with new tree */
	Lint len;		/* length of current tree */
	Lint lendash;		/* length of proposed new tree */
	Lint lenmin;		/* minimum length for any tree */
	Real r_lenmin;		/* minimum length for any tree */
	Real pacc;		/* prob. of accepting new config. */
	Branchno rootdash;	/* root of new configuration */
	Boolean newtree;	/* accepted a new configuration */
	Boolean dect;		/* should decrease temperature */
	Boolean probaccd;	/* have accepted based on Pacc */
	extern Dataptr matrix;	/* data matrix */

	/* initializations irrespective of whether restart or not */
	lengthfp = lenfp;	/* for lenlog() */
	configalloc();		/* allocate for x and xdash if necessary */
	treecopy(x, inittree);	/* current configuration */
	len = getplen(x, root);
	dect = FALSE;		/* made TRUE as necessary at end of loop */

	if (proposed == (Lint) UNSET)	/* not a restart */
	{
		lenbest = len;
		accepted = 0;
		proposed = 0;
		failedcnt = 0;
		t = t0;
		n = 0;
		iter = 0;
		bstpush(x, root);	/* initial tree initially best tree */
		if (titleprec[0] != '\0')
			lenlog("%s ", titleprec);
		lenlog("Tree\tBestLen\tTempNo\tTemp\tCurrLen\tPropLen\t"
		 "ProbAcc\n");
	}

	lenmin = getminlen(matrix);
	r_lenmin = (Real) lenmin;

	/* BEGIN BUG FIX in LVB 1.0A: enhance protection against div. by 0.0 */
	while (t > (Real) 0)	/* inf. loop might go < 0 at accuracy lim.? */
	/* END BUG FIX */
	{
		newtree = FALSE;
		probaccd = FALSE;
		rootdash = mutate(xdash, x, root);	/* propose a change */
		lendash = getplen(xdash, rootdash);
		deltalen = lendash - len;
		deltah = (r_lenmin / (Real) len) - (r_lenmin / (Real) lendash);
		if (deltalen <= 0)	/* accept the change */
		{
			if (lendash <= lenbest)		/* store tree if new */
			{
				if (lendash < lenbest)	/* very best so far */
					bstclear();	/* discard old bests */
				if (bstpush(xdash, rootdash) == 1)
					newtree = TRUE;		/* new */
			}

			/* log current stats */
			lenlog("%ld\t%ld\t%ld\t%g\t%ld\t%ld\t1\n",
			 (long) iter, (long) lenbest, (long) n, (double) t,
			 (long) len, (long) lendash);

			/* update current tree and its stats */
			len = lendash;
			treeswap(&x, &root, &xdash, &rootdash);

			if (lendash < lenbest)		/* very best so far */
				lenbest = lendash;
		}
		else			/* poss. accept change for the worse */
		{
			pacc = exp((double) (-deltah/t));
			checkedom();
			if ((Real) uni() < pacc)	/* do accept change */
			{
				probaccd = TRUE;
				treeswap(&x, &root, &xdash, &rootdash);
			}
			lenlog("%ld\t%ld\t%ld\t%g\t%ld\t%ld\t%g\n",
			 (long) iter, (long) lenbest, (long) n, (double) t,
			 (long) len, (long) lendash, (double) pacc);
			if (probaccd == TRUE)
				len = lendash;
		}
		++proposed;
		if (newtree == TRUE)
			++accepted;

		/* decide whether to reduce temperature */
		if (accepted >= maxaccept)		/* enough new trees */
		{
			failedcnt = 0;  /* this temperature a 'success' */
			dect = TRUE;
		}
		else if (proposed >= maxpropose)	/* enough proposals */
		{
			++failedcnt;
			if (failedcnt >= maxfail)	/* system frozen */
				break;			/* end of cooling */
			else				/* decrease temp. */
				dect = TRUE;
		}

		if (dect == TRUE)
		{
			++n;	/* originally n is 0 */
			t = (Real) pow((double) (t1/t0), (double) n) * t0;
			checkedom();
			proposed = 0;
			accepted = 0;
			dect = FALSE;
		}
		++iter;
		if (!(iter % CHECK_INTERVAL))
		{
			if (stopnow() == TRUE)		/* stopped by user */
				return -(len);
			if (pausenow() == TRUE)		/* paused by user */
			{
				epause(x, run, lenbest, proposed, accepted,
				 failedcnt, t, n, iter);
			}
		}
	}

	return lenbest;

}	/* end anneal() */

static void configalloc(void)
/* allocate for static x and xdash if they are NULL */
{
	if (x == NULL)
		x = treealloc();
	if (xdash == NULL)
		xdash = treealloc();
	if ((x == NULL) || (xdash == NULL))
		crash("cannot allocate for changing configurations");
	return;

}	/* end configalloc() */

Lint hillclimb(const Branch *const inittree, Branchno root,
 const Lint maxtrees, FILE *const lenfp, const char *const titleprec)
 /* find 'shortest' tree(s) using hill climbing from initial tree in inittree
 * (of root root) to look at maxtrees trees, and return the minumum length;
 * lenfp is for output of current tree length and associated details or
 * NULL for no such output;
 * if stopped prematurely, returns 0 - length */
{
        Branchno rootdash;		/* root of new config. */
        Lint deltalen;			/* change in length */
        Lint len;			/* length of current tree */
        Lint lenbest;			/* length of best tree */
        Lint lendash;			/* length of new tree */
        Lint iter; 	 		/* iter. of mut./eval. loop */

	lengthfp = lenfp;      	/* for lenlog() */
	configalloc();		/* allocate for x and xdash if necessary */
	treecopy(x, inittree);
	bstpush(x, root);	/* initial tree initially best tree */
        len = getplen(x, root);
	lenbest = len;
	if (titleprec[0] != '\0')
		lenlog("%s ", titleprec);
	lenlog("Tree\tBestLen\tCurrLen\tPropLen\n");	/* entitle columns */

	/* do maximum number of iterations */
	for (iter = 0; iter < maxtrees; ++iter)
	{
		rootdash = mutate(xdash, x, root);	/* propose change */
		lendash = getplen(xdash, rootdash);
		lenlog("%ld\t%ld\t%ld\t%ld\n", (long) iter, (long) lenbest,
		 (long) len, (long) lendash);
		deltalen = lendash - len;
		if (deltalen <= 0)	/* accept change */
		{
			treeswap(&x, &root, &xdash, &rootdash);
			len = lendash;
			if (deltalen < 0)	/* very best so far */
			{
				lenbest = len;
       				bstclear();	/* discard old bests */
       			}
	       		bstpush(x, root);
		}
		if (!(iter % CHECK_INTERVAL))
		{
			if (stopnow() == TRUE)	/* stopped by user */
				return -(len);
		}
	}

        return lenbest;

}       /* end hillclimb() */

Lint randomwalk(const Branch *const inittree, Branchno root,
 const Lint maxtrees, FILE *const lenfp, const char *const titleprec)
/* find 'shortest' tree(s) using random walk (i.e. always accept update)
 * from initial tree in inittree (of root root) to look at maxtrees trees,
 * and return the minimum length;
 * lenfp is for output of current tree length and associated details or
 * NULL for no such output;
 * if stopped prematurely, returns 0 - length */
{
	Branchno rootdash;		/* root of new config. */
	Lint len;			/* length of current tree */
	Lint lenbest;			/* length of best tree */
	Lint lendash;			/* length of new tree */
	Lint iter;			/* iter. of mut./eval. loop */
	
	lengthfp = lenfp;		/* for lenlog() */
	configalloc();
        treecopy(x, inittree);
	bstpush(x, root);	/* initial tree initially best tree */
        len = getplen(x, root);
        lenbest = len;
        if (titleprec[0] != '\0')
        	lenlog("%s ", titleprec);
	lenlog("Tree\tBestLen\tCurrLen\tProplen\n");	/* entitle columns */

	/* do requested number of iterations */
	for (iter = 0; iter < maxtrees; ++iter)
        {
                rootdash = mutate(xdash, x, root);	/* propose change */
                lendash = getplen(xdash, rootdash);
                lenlog("%ld\t%ld\t%ld%\tld\n", (long) iter, (long) lenbest,
                 (long) len, (long) lendash);
                treeswap(&x, &root, &xdash, &rootdash);		/* accept */
		len = lendash;
                if (len <= lenbest)  /* keep the tree, if new */
                {
			if (len < lenbest)	/* best so far */
                	{
				lenbest = len;
				bstclear();	/* discard old bests */
			}
			bstpush(x, root);
		}
		if (!(iter % CHECK_INTERVAL))
		{
			if (stopnow() == TRUE)
				return -(lenbest);
		}
        }

        return lenbest;

}       /* end randomwalk() */

static void checkedom(void)
/* crash if a domain error has occurred */
{
	/* BEGIN BUG FIX in LVB 1.0A: allow any integral expr. for errno */
/*	extern int errno;	*/	/* error */
	/* END BUG FIX */

	if (errno == EDOM)
		crash("domain error in solution function");

}	/* end checkedom() */

static void lenlog(const char *const fmt, ...)
/* write a message to the static length file pointer lengthfp, if it is not
 * NULL */
{
	va_list args;	/* arguments */

	if (lengthfp != NULL)
	{
		va_start(args, fmt);
		vfprintf(lengthfp, fmt, args);
		va_end(args);
	}
	return;

}	/* end lenlog() */
