/* ********** trops.c - tree operations ********** */

#include "lvb.h"

#define CLADESEP ",\n"	/* clade separator for trees */
#define DUMPROW "%d\t%d\t%d\t%d\t%d\n"	/* row format for dumped trees */

/* title for dumped tree columns */
#define DUMPTITLE "Branch:\tParent:\tLeft:\tRight:\tObject:\n"

typedef	struct	/* object set derived from a cladogram */
{
	Objno *set;	/* arrays of object sets */
	Objno cnt;	/* sizes of object sets */
}	Objset;

static Branchno arbreroot(Branch *const tree, const Branchno oldroot);
static void clear(Branch *const barray, const Branchno brnch);
static Branchno compress(Branch *const tree, const Branchno root,
 const Branchno brnch);
static void cr_bpnc(const Branch *const barray, const Branchno branch);
static void cr_chaf(const Branch *const barray, const Branchno destination,
 const Branchno newchild);
static void cr_cmpe(Branch *const barray, const Branchno branch);
static void cr_nbo(const Branch *const barray, const Objno obj);
static void cr_tbo(const Branch *const barray, const Objno obj);
static void cr_uxe(FILE *const stream, const char *const msg);
static Branchno cut(Branch *const tree, const Branchno brnch);
static void fillsets(Objset *const sstruct, const Branch *const tree,
 const Branchno root);
static void getobjs(const Branch *const barray, const Branchno root,
 Objno *const objarr, Objno *const cnt);
static Branchno getsister(const Branch *const barray, const Branchno branch);
static void lnk(Branch *const tree, const Branchno root,
 const Branchno destination, const Branchno brnch);
static Integer makesets(const Branch *const tree_1, const Branchno root_1,
 const Branch *const tree_2, const Branchno root_2);
static int objnocmp(const void *o1, const void *o2);
static int osetcmp(const void * oset1, const void *oset2);
static void r_print(FILE *const stream, const Branch *const barray,
 const Branchno root, const Boolean *const haslen);
static void randleaf(Branch *const barray, const Boolean *const leafmask,
 const Objno objs);
static void realgetobjs(const Branch *const barray, const Branchno root,
 Objno *const objarr, Objno *const cnt);
static Branchno reroot(Branch *const barray, const Branchno oldroot,
 const Branchno newroot);
static Boolean *randtopology(Branch *const barray, const Objno nobjs);
static Integer setstcmp(Objset *const oset_1, Objset *const oset_2,
 const Integer nels);
static void sort(Objset *const oset, const Integer nels);
static Objset *ssarralloc(const Integer nsets, const Integer setsize);
static void ur_print(FILE *const stream, const Branch *const barray,
 const Branchno root);

Objset *sset_1;		/* object sets for tree 1 in comparison */
Objset *sset_2;		/* object sets for tree 2 in comparison */

static void clear(Branch *const barray, const Branchno brnch)
/* initialize all scalars in branch brnch to UNSET */
{
	barray[brnch].object = UNSET;
	barray[brnch].left = UNSET;
	barray[brnch].right = UNSET;
	barray[brnch].parent = UNSET;
	return;

}	/* end clear() */

void treeclear(Branch *const barray)
/* clear all branches in array barray */
{
	extern Branchno nbranches;	/* branches in barray */
	Branchno i;			/* loop counter */

	for (i = 0; i < nbranches; ++i)
		clear(barray, i);
	return;

}	/* end treeclear() */

Branchno brcnt(void)
/* return number of branches in any binary tree structure */
{
	extern Dataptr matrix;	/* data matrix */

	return (Branchno) (matrix->n * 2 - 3);

}	/* end brcnt() */

Branchno mutate(Branch *const desttree, const Branch *const sourcetree,
 Branchno root)
/* make a copy of the tree sourcetree (of root root) in desttree,
 * with a random change in topology, the change being caused by taking a random
 * leaf branch and putting it in a random place different from where it was;
 *  return the number of the (possibly changed) root branch in the new tree */
{
	Branchno leafbranch;		/* branch to move */
	Branchno destination;		/* destination of branch to move */
	Branchno free;			/* branch freed temporarily in move */
	Branchno parent;		/* parent of branch to move */
	Branchno sister;		/* sister of branch to move */
	Branch *tree;			/* destination tree */
	extern Branchno nbranches;	/* branches in tree */

	/* for ease of reading, set up desttree with an alias, tree */
	tree = desttree;
	treecopy(tree, sourcetree);

	/* get a random leaf */
	do
	{
		leafbranch = (Branchno) randpint(nbranches - 1);
	} while (tree[leafbranch].object == UNSET);

	/* re-root tree if random leaf is the root, since can't move root */
	if (leafbranch == root)
		root = arbreroot(tree, root);

	parent = tree[leafbranch].parent;
	if (parent == UNSET)
		crash("internal error in function mutate(): parent of random\n"
		 "leaf (%d) is UNSET", (int) leafbranch);
	sister = getsister(tree, leafbranch);
	if (sister == UNSET)
		crash("internal error in function mutate(): sister of random\n"
		 "leaf (%d) is UNSET", (int) leafbranch);

	/* get a destination that is not source or its parent or sister */
	do
	{
		destination = (Branchno) randpint(nbranches - 1);
	} while ((destination == leafbranch) || (destination == parent)
	 || (destination == sister));

	parent = cut(tree, leafbranch);		/* excise source branch */
	free = compress(tree, root, parent);	/* make tree intact */
	lnk(tree, root, destination, free);	/* make space for new child */
	childadd(tree, destination, leafbranch);
	return root;

}	/* end mutate() */

static Branchno arbreroot(Branch *const tree, const Branchno oldroot)
/* change tree's root arbitrarily, to a leaf other than oldroot;
 * return the number of the new root */
{
	Branchno newroot = 0;		/* new root */

	/* find a leaf that is not the current root */
	while ((tree[newroot].object == UNSET) || (newroot == oldroot))
		++newroot;

	reroot(tree, oldroot, newroot);
	return newroot;

}	/* end arbreroot() */

static Branchno reroot(Branch *const barray, const Branchno oldroot,
 const Branchno newroot)
/* change the root of the tree in barray from oldroot to newroot, which must
 * not be the same;
 * return oldroot */
{
	Branchno current;			/* current branch */
	Branchno parnt;				/* parent of current branch */
	Branchno sister = UNSET;		/* sister of current branch */
	Branchno previous;			/* previous branch */
	static Branchno *oldparent = NULL;	/* el. i is old parent of i */
	extern Branchno nbranches;		/* branches in tree */

	/* check new root is a leaf but not the current root */
	if (barray[newroot].object == UNSET)
		crash("internal error in function reroot(): proposed new\n"
		 "root (branch %d) is interior\n", (int) newroot);
	if ((newroot == oldroot))
		crash("internal error in function reroot(): old root and\n"
		 "proposed new root are the same (branch %d)\n",
		 (int) oldroot);

	/* create record of parents as they are now */
	if (oldparent == NULL)	/* 1st call: allocate memory */
	{
		oldparent = (Branchno *)
		 alloc((size_t) nbranches * sizeof(Branchno),
		 "old parent array");
		mempush(oldparent);
	}
	for (current = 0; current < nbranches; ++current)
		oldparent[current] = barray[current].parent;

	current = newroot;
	previous = UNSET;
	while (current != oldroot)
	{
		if (current == UNSET)
			crash("internal error in function reroot(): current\n"
			 "branch is UNSET");
		parnt = oldparent[current];		/* original parent */
		if (current == barray[parnt].left)
			sister = barray[parnt].right;
		else if (current == barray[parnt].right)
			sister = barray[parnt].left;
		else	/* error in tree structure */
			crash("internal error in function reroot(): current\n"
			 "branch %d has old parent %d, but old parent does\n"
			  "not have it as a child", (int) current,
			  (int) parnt);
		barray[current].parent = previous;	/* now chld of prev. */

		/* make former parent the new left child, and former sister the
		 new right child of the current branch */
		barray[current].left = parnt;
		barray[current].right = sister;
		barray[parnt].parent = current;
		barray[sister].parent = current;
		
		/* move towards orig. root, i.e. to orig. parnt of curr. br. */
		previous = current;
		current = parnt;
	}

	/* former root is now a normal leaf, without descendants */
	barray[oldroot].left = UNSET;
	barray[oldroot].right = UNSET;

	return oldroot;

}	/* end reroot() */

Branchno objreroot(Branch *const barray, const Branchno oldroot,
 const Objno newrobj)
/* change the root of the tree in barray from oldroot to the branch associated
 * with object newrobj;
 * return the number of the new root */
{
	Branchno i;			/* loop counter */
	Branchno newroot = UNSET;	/* new root */
	extern Branchno nbranches;	/* number of branches in tree */
	Boolean foundroot = FALSE;	/* have found new root object */

	for (i = 0; i < nbranches; ++i)
	{
		if (barray[i].object == newrobj)
		{
			if (foundroot == TRUE)
				cr_tbo(barray, newrobj);
			newroot = i;
			foundroot = TRUE;
		}
	}
	if (newroot == UNSET)
		cr_nbo(barray, newrobj);
	else if (newroot != oldroot)
		reroot(barray, oldroot, newroot);
	return newroot;

}	/* end objreroot() */

static void cr_nbo(const Branch *const barray, const Objno obj) 
/* crash because no branches of tree in barray have proposed root object obj
 * as their object */
{
	crash("internal error: in tree array %p, no branch is associated\n"
	 "with proposed root object %d", (void *) barray, (int) obj);

}	/* end cr_nbo() */

static void cr_tbo(const Branch *const barray, const Objno obj)
/* crash because 2 branches of tree in barray have object obj */
{
	crash("internal error in tree array %p, 2 branch records claim\n"
	 "object %d", (void *) barray, (int) obj);
	
}	/* end cr_tbo() */

static Branchno getsister(const Branch *const barray, const Branchno branch)
/* return number of sister of branch branch in tree in barray, or UNSET if
 * branch has none */
{
	Branchno parnt;		/* parent of current branch */

	parnt = barray[branch].parent;
	if (parnt == UNSET)
		return UNSET;
	if (branch == barray[parnt].left)
		return barray[parnt].right;
	else if (branch == barray[parnt].right)
		return barray[parnt].left;
	else	/* error in tree structure */
	{
		cr_bpnc(barray, branch);
		return 0;	/* NEVER reached but it shuts up compilers */
	}

}	/* end getsister() */

Branchno getroot(const Branch *const barray)
/* return index of root branch in tree in barray */
{
	Branchno i;			/* loop counter */
	extern Branchno nbranches;	/* branches in tree */

	for (i = 0; i < nbranches; ++i)
	{
		if ((barray[i].parent == UNSET) && (barray[i].object != UNSET))
			return i;
	}
	crash("internal error in tree array %p of %d branches:\n"
	 "there is no root branch", (void *) barray, (int) nbranches);
	return 0;	/* NEVER reached but it shuts up compilers */
}	/* end getroot() */

Branchno childadd(Branch *const tree, const Branchno destination,
 const Branchno newchild)
/* replace unset child of destination with newchild, and return destination */
{
	if (tree[destination].right == UNSET)
		tree[destination].right = newchild;
	else if (tree[destination].left == UNSET)
		tree[destination].left = newchild;
	else	/* error: destination already has 2 children */
		cr_chaf(tree, destination, newchild);
	tree[newchild].parent = destination;
	return destination;

}	/* end childadd() */

static void cr_chaf(const Branch *const barray, const Branchno destination, 
 const Branchno newchild)
/* crash because we want to add branch newchild to the children of branch
 * destination in tree in barray, but it already has 2 so there is no room */
{
	crash("internal error in tree array %p: cannot make branch %d a\n"
	 "child of branch %d since this already has 2 children (left is\n"
	 "branch %d, right is branch %d)", (void *) barray, (int) newchild,
	 (int) destination, (int) barray[destination].left,
	 (int) barray[destination].right);

}	/* end cr_chaf() */

static void lnk(Branch *const tree, const Branchno root,
 const Branchno destination, const Branchno brnch)
/* link brnch into the tree tree (of root root) by adding it between
 * destination and destination's children */
{
	Branchno lft;	/* left child of destination */
	Branchno rght;	/* right child of destination */

	lft = tree[destination].left;
	rght = tree[destination].right;

	tree[brnch].left = lft;
	tree[brnch].right = rght;
	tree[brnch].parent = destination;

	/* make brnch the only child of dest.; left is chosen arbitrarily */
	tree[destination].left = brnch;
	tree[destination].right = UNSET;

	/* give destination's children brnch as their parent, if they exist */
	if (lft != UNSET)
		tree[lft].parent = brnch;
	if(rght != UNSET)
		tree[rght].parent = brnch;

	/* move dest's object, if any, to brnch, unless destination is root */
	if (destination == root)
		tree[brnch].object = UNSET;
	else
	{
		tree[brnch].object = tree[destination].object;
		tree[destination].object = UNSET;
	}

	return;

}	/* end lnk () */

static Branchno compress(Branch *const tree, const Branchno root,
 const Branchno brnch)
/* collapse the branch leading from brnch, assuming there is only one such;
 * return the number of the freed branch */
{
	Branchno child = UNSET;		/* live child of brnch */
	Branchno lft;			/* left child of child */
	Branchno rght;			/* right child of child */

	/* find live child, or crash if there are 2 */
	if (tree[brnch].left != UNSET)
		child = tree[brnch].left;
	else if (tree[brnch].right != UNSET)
		child = tree[brnch].right;
	else	/* error: branch has two children */
		cr_cmpe(tree, brnch);

	lft = tree[child].left;
	rght = tree[child].right;
	tree[brnch].left = lft;
	tree[brnch].right = rght;
	if (lft != UNSET)
		tree[lft].parent = brnch;
	if (rght != UNSET)
		tree[rght].parent = brnch;

	/* give obj. of freed child, unless root (would UNSET root's object) */
	if (brnch != root)
		tree[brnch].object = tree[child].object;

	clear(tree, child);
	return child;

}	/* end compress() */

static void cr_cmpe(Branch *const barray, const Branchno branch)
/* crash because branch branch in array barray has 2 children, but 1 was
 * expected [by compress()] */
{
	crash("internal error in tree array %p: branch %d has 2 children\n"
	 "when 1 was expected\n", (void *) barray, (int) branch);

}	/* end cr_cmpe() */

static Branchno cut(Branch *const tree, const Branchno brnch)
/* excise branch brnch from tree;
return former parent of excised branch */
{
	Branchno parnt;		/* parent of branch to cut */

	parnt = tree[brnch].parent;
	if (tree[parnt].left == brnch)
		tree[parnt].left = UNSET;
	else if (tree[parnt].right == brnch)
		tree[parnt].right = UNSET;
	else	/* error in tree structure: branch and parent not connected */
		cr_bpnc(tree, brnch);

	tree[brnch].parent = UNSET;
	return parnt;

}	/* end cut() */

static void cr_bpnc(const Branch *const barray, const Branchno branch)
/* crash because branch branch in tree in barray is not connected to its
 * parent, i.e. it is not a child of the branch it claims as parent, according
 * to that 'parent's' record of its own children */
{
	const Branchno parnt = barray[branch].parent;	/* parent of branch */

	crash("internal error in tree array %p: branch record %d says it has\n"
	 "parent %d, but branch record %d has left child %d and right child "
	 "%d", (void *) barray, (int) branch, (int) parnt, (int) parnt,
	 (int) barray[parnt].left, (int) barray[parnt].right);

}	/* end cr_bpnc() */

void treecopy(Branch *const dest, const Branch *const source)
/* copy tree from source to dest;
dest must have enough space */
{
	extern Branchno nbranches;	/* branches in tree */

	memcpy(dest, source, (size_t) nbranches * sizeof(Branch));
	return;

}	/* end treecopy() */

void randtree(Branch *const barray)
/* fill barray with a random tree, where barray[0] is the root */
{
	Boolean *leafmask;	/* TRUE where branch in array is a leaf */
	extern Dataptr matrix;	/* data matrix */

	treeclear(barray);
	leafmask = randtopology(barray, matrix->n);
	randleaf(barray, leafmask, matrix->n);
	return;

}	/* end randtree() */

Branch *treealloc(void)
/* return array of extern nbranches branches with scalars all UNSET, or NULL if
 * impossible;
 * N.B. pushes array on to the general memstack */
{
	Branch *barray;			/* tree */
	extern Branchno nbranches;	/* branches per tree */

	barray = (Branch *) malloc((size_t) nbranches * sizeof(Branch));
	if (barray != NULL)
	{
		mempush(barray);
		treeclear(barray);
	}
	return barray;

}	/* end treealloc() */

static Boolean *randtopology(Branch *const barray, const Objno nobjs)
/* fill barray with tree of random topology, where barray[0] is root;
 * return static array where element i is TRUE if barray[i] is a leaf,
 * or, FALSE if it is not;
 * this array will be overwritten in subsequent calls and is pushed on to the
 * general memstack */
{
	Branchno i;			/* loop counter */
	Branchno leaves = 0;		/* number of leaves */
	Branchno nextfree = 0;		/* next unused el. of barray */
	Branchno togrow;		/* random candidate for sprouting */
	static Boolean *isleaf = NULL;	/* return value */
	extern Branchno nbranches;	/* branches in tree */

	if (isleaf == NULL)	/* 1st call, allocate memory */
	{
		isleaf = (Boolean *)
		 alloc((size_t) nbranches * sizeof(Boolean), "leaf mask");
		mempush(isleaf);
	}

	/* clear the leaf mask */
	for (i = 0; i < nbranches; ++i)
		isleaf[i] = FALSE;

	/* start with initial tree of 3 leaves */
	barray[0].parent = UNSET;
	isleaf[nextfree++] = TRUE;
	barray[0].left = nextfree;
	barray[nextfree].parent = 0;
	isleaf[nextfree++] = TRUE;
	barray[0].right = nextfree;
	barray[nextfree].parent = 0;
	isleaf[nextfree++] = TRUE;
	leaves = 3;

	/* sprout! */
	while(leaves < nobjs)
	{
		do	/* select a random leaf other than the root */
		{
			togrow = 1 + (Branchno) randpint(nextfree - 2);
		} while (isleaf[togrow] == FALSE);
		/* left child */
		barray[togrow].left = nextfree;
		barray[nextfree].parent = togrow;
		isleaf[nextfree++] = TRUE;
		/* right child */
		barray[togrow].right = nextfree;
		barray[nextfree].parent = togrow;
		isleaf[nextfree++] = TRUE;
		/* other updates */
		isleaf[togrow] = FALSE;
		++leaves;
	}

	return isleaf;

}	/* end randtopology() */

static void randleaf(Branch *const barray, const Boolean *const leafmask,
 const Objno objs)
/* randomly assign objects numbered 0 to objs - 1 to leaves of tree in barray;
leaves in barray must be indicated by corresponding TRUEs in leafmask */
{
	Objno assigned = 0;		/* for safety: should == objs at end */
	Objno candidate;		/* random object */
	Branchno i;			/* loop counter */
	static Boolean *used = NULL;	/* el. i TRUE if obj. i has leaf */
	extern Branchno nbranches;	/* number of branches in tree */

	if (used == NULL)		/* 1st call, allocate memory */
	{
		used = (Boolean *) alloc((size_t) objs * sizeof(Boolean),
		 "used object mask");
		mempush(used);
	}

	/* clear 'used' array */
	for (i = 0; i < objs; ++i)
		used[i] = FALSE;

	/* assign an object to every leaf */
	for (i = 0; i < nbranches; ++i)
	{
		if (leafmask[i] == TRUE)	/* leaf, requires object */
		{
			do	/* get a new object number */
			{
				candidate = randpint(objs - 1);
			} while(used[candidate] == TRUE);
			/* assign object to leaf */
			barray[i].object = candidate;
			used[candidate] = TRUE;
			++assigned;
		}
	}

	/* for safety, check on number of objects assigned */
	if (assigned != objs)
		crash("internal error in function randleaf(): %d objects\n"
		 "exist, but have assigned %d objects to branches", (int) objs,
		 (int) assigned);

	return;

}	/* end randleaf() */

void treeswap(Branch **const tree1, Branchno *const root1,
 Branch **const tree2, Branchno *const root2)
/* swap trees pointed to by tree1 and tree2;
 * also swap records of their roots, as pointed to by root1 and root2 */
{
	Branchno tmproot;	/* temporary value-holder for swapping roots */
	Branch *tmptree;	/* temporary value-holder for swapping trees */

	/* swap pointers to branch arrays */
	tmptree = *tree1;
	*tree1 = *tree2;
	*tree2 = tmptree;

	/* swap roots */
	tmproot = *root1;
	*root1 = *root2;
	*root2 = tmproot;

	return;

}	/* end treeswap() */

void treedump(FILE *const stream, const Branch *const tree)
/* send tree as table of integers to file pointed to by stream */
{
	Branchno i;			/* loop counter */
	Objno obj;			/* obj. assoc'd with current branch */
	extern Branchno nbranches;	/* branches in tree */

	fprintf(stream, DUMPTITLE);
	for (i = 0; i < nbranches; ++i)
	{
		obj = tree[i].object;
		fprintf(stream, DUMPROW, (int) i, (int) tree[i].parent,
		 (int) tree[i].left, (int) tree[i].right, (int) obj);
	}
	putc('\n', stream);
	if(ferror(stream) != 0)
		cr_uxe(stream, "dumping tree");
	return;

}	/* end treedump() */

static void cr_uxe(FILE *const stream, const char *const msg)
/* crash because of problem on file stream, with message consisting of
 * "FATAL ERROR: ", then "file error " if the error indicator for stream is set
 * or "unexpected end of file " if not, then "when ", followed by msg */
{
	if (ferror(stream))
		crash("file error when %s", msg);
	else
		crash("unexpected end of file when %s", msg);

}	/* end cr_uxe */

void trundump(FILE *const stream, Branch *const barray)
/* read the dumped tree at the current position in stream into barray,
 * rerooting where necessary to ensure barray[0] is the root */
{
	const char act[] = "reading "
	 "dumped tree file";		/* activity str. for crash messages */
	int i = 0;			/* current branch */
	int inchar;			/* character of input */
	int par;			/* parent of current branch */
	int lft;			/* left child of current branch */
	int rght;			/* right child of current branch */
	int obj;			/* obj. associated with curr. branch */
	Branchno oldroot;		/* root of dumped tree */
	extern Branchno nbranches;	/* branches per tree */

	/* move past column titles */
	while ((inchar = getc(stream)) != '0')	/* go to first branch number */
	{
		if (inchar == EOF)
			cr_uxe(stream, act);
	}
	if (ungetc('0', stream) != '0')
		crash("file error when %s", act);

	/* get tree */
	while (i < (int) (nbranches - 1))
	{
		if (fscanf(stream, DUMPROW, &i, &par, &lft, &rght, &obj) != 5)
			crash("error %s", act);
		barray[i].parent = (Branchno) par;
		barray[i].left = (Branchno) lft;
		barray[i].right = (Branchno) rght;
		barray[i].object = (Objno) obj;
	}
	if (ferror(stream) != 0)
		cr_uxe(stream, act);
	
	/* re-root tree if necessary */
	oldroot = getroot(barray);
	if (oldroot != 0)
	/* BEGIN BUG FIX in LVB 1.0: allow original branch 0 to be internal */
	{
		if (barray[0].object != UNSET)	/* 0 is a leaf, can reroot() */
			reroot(barray, oldroot, 0);
		else				/* 0 intern., can't reroot() */
		{
			Branch tmp;	/* temporary branch for swapping */
			Branchno br;	/* loop counter */
			for (br = 0; br < nbranches; ++br)
			{
				if (barray[br].left == 0)
					barray[br].left = oldroot;
				if (barray[br].right == 0)
					barray[br].right = oldroot;
				if (barray[br].parent == 0)
					barray[br].parent = oldroot;
				else if (barray[br].parent == oldroot)
					barray[br].parent = 0;
			}
			tmp = barray[0];
			barray[0] = barray[oldroot];
			barray[oldroot] = tmp;
		}
	}
	/* END BUG FIX */

	return;

}	/* end trundump() */

void treeprint (FILE *const stream, const Branch *const barray,
 const Branchno root, const Boolean *const haslen)
/* print tree in barray (of root root) in bracketed text form to stream stream;
 * if haslen is NULL, the tree is output unrooted;
 * otherwise, haslen is an array FALSE where branches are to be collapsed and
 * TRUE elsewhere */
{
	if (haslen == NULL)
		ur_print(stream, barray, root);
	else
		r_print(stream, barray, root, haslen);
	return;

}	/* end treeprint() */

static void ur_print(FILE *const stream, const Branch *const barray,
 const Branchno root)
/* send tree in barray, of root root, to file pointed to by stream in
 * unrooted form */
{
	Branchno obj;				/* current object */
	static Boolean doneabsroot = FALSE;	/* have output root of tree */
	static Boolean usecomma;		/* do output clade separator */
	extern Dataptr matrix;			/* data matrix */

	obj = barray[root].object;

	if (doneabsroot == FALSE)	/* print whole tree */
	{
		/* start tree */
		fprintf(stream, "(%s", matrix->rowtitle[obj]);
		usecomma = TRUE;
		doneabsroot = TRUE;

		ur_print(stream, barray, barray[root].left);
		ur_print(stream, barray, barray[root].right);

		/* end tree */
		fputs(");\n", stream);
		if (ferror(stream))
			crash("file error when writing unrooted tree");

		/* clean up for next call */
		usecomma = FALSE;
		doneabsroot = FALSE;
	}
	else	/* print remainder of tree */
	{
		if (usecomma == TRUE)
			fputs(CLADESEP, stream);
		if (barray[root].object != UNSET)	/* leaf */
		{
			fprintf(stream, "%s", matrix->rowtitle[obj]);
			usecomma = TRUE;
		}
		else
		{
			putc('(', stream);
			usecomma = FALSE;
			ur_print(stream, barray, barray[root].left);
			ur_print(stream, barray, barray[root].right);
			fputc(')', stream);
			usecomma = TRUE;
		}
	}
	return;

}	/* end ur_print() */

static void r_print(FILE *const stream, const Branch *const barray,
 const Branchno root, const Boolean *const haslen)
/* send tree in barray, of root root, to file pointed to by stream in
 * rooted form;
 * collapse ingroup branches where the corresponding element of haslen is
 * FALSE */
{
        Branchno obj;                           /* current object */
        static Boolean doneabsroot = FALSE;     /* have output root of tree */
        static Boolean usecomma;                /* should output a comma */
        extern Dataptr matrix;                  /* data matrix */
 
        obj = barray[root].object;
        
        if (doneabsroot == FALSE)       /* print whole tree */
        {   
                /* start tree */
                fprintf(stream, "(%s%s(", matrix->rowtitle[obj], CLADESEP);
                usecomma = FALSE;
                doneabsroot = TRUE;
 
                treeprint(stream, barray, barray[root].left, haslen);
                treeprint(stream, barray, barray[root].right, haslen);
 
                /* end tree */
                fputs("));\n", stream);
                if (ferror(stream))
                        crash("file error when writing rooted tree");
        
                /* clean up for next call */
                usecomma = FALSE;
                doneabsroot = FALSE;
        }
        else    /* print remainder of tree */
        {
		if (usecomma == TRUE)
			fputs(CLADESEP, stream);       
                if (barray[root].object != UNSET)       /* leaf */
                {       
                        fprintf(stream, "%s", matrix->rowtitle[obj]);
                        usecomma = TRUE;
                }
                else
                {
			if (haslen[root])	/* start new clade */
				putc('(', stream);
			usecomma = FALSE;	/* did ',' after prev. name */
                        treeprint(stream, barray, barray[root].left, haslen);
                        treeprint(stream, barray, barray[root].right, haslen);
			if (haslen[root])	/* terminate clade */
			{
                                fputc(')', stream);
                                usecomma = TRUE;
			}
                }
        }
        return;
 
}       /* end r_print() */

Integer treecmp(const Branch *const tree_1, const Branchno root_1,
 const Branch *const tree_2, Branchno root_2)
/* return 0 if the topology of tree_1 (of root root_1) is the same as that
 * of tree_2 (of root root_2), or non-zero if different */
{
	static Branch *copy_2 = NULL;	/* possibly re-rooted tree 2 */	
	Integer nsets;			/* elements per set array */

	/* set up local copy of tree 2 with same root as tree 1 */
	if (copy_2 == NULL)	/* first call, allocate memory */
	{
		copy_2 = treealloc();
		if (copy_2 == NULL)
			crash("out of memory: cannot allocate for copy of\n"
			 "tree for comparison");
	}
	treecopy(copy_2, tree_2);
	root_2 = objreroot(copy_2, root_2, tree_1[root_1].object);

	nsets = makesets(tree_1, root_1, copy_2, root_2);
	return setstcmp(sset_1, sset_2, nsets);

}	/* end treecmp() */

static Integer setstcmp(Objset *const oset_1, Objset *const oset_2,
 const Integer nels)
/* return 0 if the same sets of objects are in oset_1 and oset_2,
 * and non-zero otherwise */
{
	Integer i;	/* loop counter */

	/* sort the set arrays and their constituent sets */
	sort(oset_1, nels);
	sort(oset_2, nels);

	/* compare the set arrays */
	for (i = 0; i < nels; ++i)
	{
		if (oset_1[i].cnt != oset_2[i].cnt)
			return 1;
		else if (memcmp(oset_1[i].set, oset_2[i].set,
		 (size_t) oset_1[i].cnt * sizeof(Objno)) != 0)
		 	return 1;
	}

	return 0;

}	/* end setstcmp() */

static void sort(Objset *const oset, const Integer nels)
/* sort the nels object sets in oset so that each is in order, and sort oset so
 * that the sets themselves are in order of size and content */
{
	Integer i;	/* loop counter */

	/* first sort each set member list */
	for (i = 0; i < nels; ++i)
		qsort(oset[i].set, (size_t) oset[i].cnt, sizeof(Objno),
		 objnocmp);

	/* now sort the arrays of sets by size and content */
	for (i = 0; i < nels; ++i)
		qsort(oset, (size_t) nels, sizeof(Objset), osetcmp);

	return;

 }	/* end sort() */

static int osetcmp(const void *oset1, const void *oset2)
/* comparison function for object sets (type Objset):
 * return negative if *oset1 is a smaller set of objects than *oset2 or is the
 * same size but with a list of elements that compares lower;
 * return positive if *ostet1 is bigger or the same size but with a list of
 * elements that compares higher;
 * return 0 if they are the same;
 * N.B. the object numbers must be in numerical order within the sets */
{
	Integer i;					/* loop counter */
	const Objset loset_1 = *((Objset *) oset1);	/* simpler alias */
	const Objset loset_2 = *((Objset *) oset2);	/* simpler alias */

	/* sets of different size differ */
	if (loset_1.cnt != loset_2.cnt)
		return (int) (loset_1.cnt - loset_2.cnt);

	/* then see if sets' contents differ */
	for (i = 0; i < loset_1.cnt; ++i)
	{
		if (loset_1.set[i] != loset_2.set[i])
			return (int) (loset_1.set[i] - loset_2.set[i]);
	}

	return 0;	/* really they are the same */

}	/* end osetcmp() */

static Integer makesets(const Branch *const tree_1, const Branchno root_1,
 const Branch *const tree_2, const Branchno root_2)
/* fill static sset_1 and static sset_2 with arrays of object sets for
 * tree_1 and tree_2 (of root_1 and root_2 respectively), and return the
 * extent of each array;
 * the trees must have the same object in the root branch;
 * arrays will be overwritten on subsequent calls */
{
	extern Dataptr matrix;				/* data matrix */
	const Integer nsets = (Integer) matrix->n - 3;	/* sets per tree */
	const Integer mssz = (Integer) matrix->n - 2;	/* max. objs per set */

	if (sset_1 == NULL)	/* first call, allocate memory */
	{
		sset_1 = ssarralloc(nsets, mssz);
		sset_2 = ssarralloc(nsets, mssz);
	}
	fillsets(sset_1, tree_1, root_1);
	fillsets(sset_2, tree_2, root_2);
	return nsets;	/* set by ssarralloc() */

}	/* end makesets() */

static Objset *ssarralloc(const Integer nsets, const Integer setsize)
/* return a new object set array with space for nsets object sets of setsize
 * objects each;
 * new memory is pushed on to the general memstack */
{
	Objset *nobjset;				/* new set array */
	Integer i;					/* loop counter */

	nobjset = (Objset *)
	 alloc((size_t) nsets * sizeof(Objset), "object set struct array");
	mempush(nobjset);
	for (i = 0; i < nsets; ++i)
	{
		nobjset[i].set = (Objno *) alloc((size_t)
		 setsize * sizeof(Objno), "object set object arrays");
		mempush(nobjset[i].set);
		nobjset[i].cnt = UNSET;
	}
	return nobjset;

}	/* end ssarralloc() */

static void fillsets(Objset *const sstruct, const Branch *const tree,
 const Branchno root)
/* fill object sets in sstruct with all sets of objects in tree tree, descended
 * from but not including root and not including sets of one object */
{
	static Integer i = UNSET;		/* current set being filled */

	if (i == UNSET)		/* not a recursive call */
	{
		i = 0;

		/* avoid generating sets for true root and leaves */
		if (tree[tree[root].left].object == UNSET)	/* interior */
			fillsets(sstruct, tree, tree[root].left);
		if (tree[tree[root].right].object == UNSET)	/* interior */
		fillsets(sstruct, tree, tree[root].right);

		i = UNSET;	/* clean up for next non-recursive call */
		return;
	}
	if (tree[root].left != UNSET)	/* not leaf */
	{
		getobjs(tree, root, sstruct[i].set, &sstruct[i].cnt);
		++i;
		fillsets(sstruct, tree, tree[root].left);
		fillsets(sstruct, tree, tree[root].right);
		return;
	}

}	/* end fillsets */

static void getobjs(const Branch *const barray, const Branchno root,
 Objno *const objarr, Objno *const cnt)
/* fill objarr (which must be large enough) with numbers of all objects
 * in the tree in barray in the clade starting at branch root;
 * fill the number pointed to by cnt with the number of objects found
 * (i.e. the number of elements written to objarr) */
{
	*cnt = 0;
	realgetobjs(barray, root, objarr, cnt);

}	/* end getobjs() */

static void realgetobjs(const Branch *const barray, const Branchno root,
 Objno *const objarr, Objno *const cnt)
/* fill objarr (which must be large enough) with numbers of all objects
 * in the tree in barray in the clade starting at branch root;
 * fill the number pointed to by cnt, which must initially be zero, with
 * the number of objects found (i.e. the number of elements written to objarr);
 * this function should not be called from anywhere except getobjs(), which is
 * a safer interface */
{
	if (barray[root].object != UNSET)
	{
		objarr[*cnt] = barray[root].object;
		++(*cnt);
	}
	else
	{
		if (barray[root].left != UNSET)
			realgetobjs(barray, barray[root].left, objarr, cnt);
		if (barray[root].right != UNSET)
			realgetobjs(barray, barray[root].right, objarr, cnt);
	}
	return;

}	/* end realgetobjs() */

static int objnocmp(const void *o1, const void *o2)
/* comparison function for comparing object numbers:
 * return negative if *o1 < *o2, zero if *o1 == *o2,
 * or positive if *o1 > *o2 */
{

	return((int) (*((Objno *) o1) - *((Objno *) o2)));

}	/* end objnocmp() */
